#!/usr/bin/env python3

# ***** BEGIN GPL LICENSE BLOCK ***** #
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Contributor(s): Campbell Barton
#
# ***** END GPL LICENSE BLOCK *****

# <pep8 compliant>

"""
This is a tool for reviewing commit ranges, writing into accept/reject files,
and optionally generate release-log-ready data.

Useful for reviewing revisions to backport to stable builds.

Note that, if any of the data files generated already exist, they will be extended
with new revisions, not overwritten.

Note that, for the most complex 'wiki-ready' file generated by  `--accept-releaselog`,
proof-reading after this tool has ran is heavily suggested!

Example usage:

   ./git_log_review_commits_advanced.py  --source ../../.. --range HEAD~40..HEAD --filter 'BUGFIX' --accept-pretty --accept-releaselog --blender-rev 2.79

To add list of fixes between RC2 and RC3, and list both RC2 and RC3 fixes also in their own sections:

   ./git_log_review_commits_advanced.py  --source ../../.. --range <RC2 revision>..<RC3 revision> --filter 'BUGFIX' --accept-pretty --accept-releaselog --blender-rev 2.79 --blender-rstate=RC3 --blender-rstate-list="RC2,RC3"

To exclude all commits from some given files, by sha1 or by commit message (from previously generated release logs) - much handy when going over commits which were partially cherry-picked into a previous release branch e.g.:

   ./git_log_review_commits_advanced.py  --source ../../.. --range HEAD~40..HEAD --filter 'BUGFIX' --filter-exclude-sha1-fromfiles "review_accept.txt" "review_reject.txt" --filter-exclude-fromreleaselogs "review_accept_release_log.txt" --accept-pretty --accept-releaselog --blender-rev 2.75

"""

import os
import sys
import io
import re

ACCEPT_FILE = "review_accept.txt"
REJECT_FILE = "review_reject.txt"
ACCEPT_LOG_FILE = "review_accept_log.txt"
ACCEPT_PRETTY_FILE = "review_accept_pretty.txt"
ACCEPT_RELEASELOG_FILE = "review_accept_release_log.txt"

IGNORE_START_LINE = "<!-- IGNORE_START -->"
IGNORE_END_LINE = "<!-- IGNORE_END -->"

_cwd = os.getcwd()
__doc__ = __doc__ + \
    "\nRaw GIT revisions files:\n\t* Accepted: %s\n\t* Rejected: %s\n\n" \
    "Basic log accepted revisions: %s\n\nWiki-printed accepted revisions: %s\n\n" \
    "Full release notes wiki page: %s\n" \
    % (os.path.join(_cwd, ACCEPT_FILE), os.path.join(_cwd, REJECT_FILE),
       os.path.join(_cwd, ACCEPT_LOG_FILE), os.path.join(_cwd, ACCEPT_PRETTY_FILE),
       os.path.join(_cwd, ACCEPT_RELEASELOG_FILE))
del _cwd


class _Getch:
    """
    Gets a single character from standard input.
    Does not echo to the screen.
    """

    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self):
        return self.impl()


class _GetchUnix:

    def __init__(self):
        import tty
        import sys

    def __call__(self):
        import sys
        import tty
        import termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:

    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()


getch = _Getch()
# ------------------------------------------------------------------------------
# Pretty Printing

USE_COLOR = True

if USE_COLOR:
    color_codes = {
        'black': '\033[0;30m',
        'bright_gray': '\033[0;37m',
        'blue': '\033[0;34m',
        'white': '\033[1;37m',
        'green': '\033[0;32m',
        'bright_blue': '\033[1;34m',
        'cyan': '\033[0;36m',
        'bright_green': '\033[1;32m',
        'red': '\033[0;31m',
        'bright_cyan': '\033[1;36m',
        'purple': '\033[0;35m',
        'bright_red': '\033[1;31m',
        'yellow': '\033[0;33m',
        'bright_purple': '\033[1;35m',
        'dark_gray': '\033[1;30m',
        'bright_yellow': '\033[1;33m',
        'normal': '\033[0m',
    }

    def colorize(msg, color=None):
        return (color_codes[color] + msg + color_codes['normal'])
else:
    def colorize(msg, color=None):
        return msg
bugfix = ""


BUGFIX_CATEGORIES = (
    ("Objects / Animation / GP", (
        "Animation",
        "Constraints",
        "Grease Pencil",
        "Objects",
        "Dependency Graph",
    ),
    ),

    ("Data / Geometry", (
        "Armatures",
        "Curve/Text Editing",
        "Mesh Editing",
        "Meta Editing",
        "Modifiers",
        "Material / Texture",
    ),
    ),

    ("Physics / Simulations / Sculpt / Paint", (
        "Particles",
        "Physics / Hair / Simulations",
        "Sculpting / Painting",
    ),
    ),

    ("Image / Video / Render", (
        "Image / UV Editing",
        "Masking",
        "Motion Tracking",
        "Movie Clip Editor",
        "Nodes / Compositor",
        "Render",
        "Render: Cycles",
        "Render: Freestyle",
        "Sequencer",
    ),
    ),

    ("UI / Spaces / Transform", (
        "3D View",
        "Input (NDOF / 3D Mouse)",
        "Outliner",
        "Text Editor",
        "Transform",
        "User Interface",
    ),
    ),

    ("Game Engine", (
    ),
    ),

    ("System / Misc", (
        "Audio",
        "Collada",
        "File I/O",
        "Other",
        "Python",
        "System",
    ),
    ),
)


sys.stdin = os.fdopen(sys.stdin.fileno(), "rb")
sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8', errors='surrogateescape', line_buffering=True)
sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding='utf-8', errors='surrogateescape', line_buffering=True)


def gen_commit_summary(c):
    # In git, all commit message lines until first empty one are part of 'summary'.
    return c.body.split("\n\n")[0].strip(" :.;-\n").replace("\n", " ")


def print_commit(c):
    print("------------------------------------------------------------------------------")
    print(colorize(c.sha1.decode(), color='green'), end=" ")
    print(colorize(c.date.strftime("%Y/%m/%d"), color='purple'), end=" ")
    print(colorize(c.author, color='bright_blue'))
    print()
    print(colorize(c.body, color='normal'))
    print()
    print(colorize("Files: (%d)" % len(c.files_status), color='yellow'))
    for f in c.files_status:
        print(colorize("  %s %s" % (f[0].decode('ascii'), f[1].decode('ascii')), 'yellow'))
    print()


def gen_commit_log(c):
    return "rB%s   %s   %-30s   %s" % (c.sha1.decode()[:10], c.date.strftime("%Y/%m/%d"),
                                       c.author, gen_commit_summary(c))


re_bugify_str = r"T([0-9]{1,})"
re_bugify = re.compile(re_bugify_str)
re_commitify = re.compile(r"\W(r(?:B|BA|BAC|BTS)[0-9a-fA-F]{6,})")
re_prettify = re.compile(r"(.{,20}?)(Fix(?:ing|es)?\s*(?:for)?\s*" + re_bugify_str + r")\s*[-:,]*\s*", re.IGNORECASE)


def gen_commit_message_pretty(c, unreported=None):
    body = gen_commit_summary(c)

    tbody = re_prettify.sub(r"Fix {{BugReport|\3}}: \1", body)
    if tbody == body:
        if unreported is not None:
            unreported[0] = True
        tbody = "Fix unreported: %s" % body
    body = re_bugify.sub(r"{{BugReport|\1}}", tbody)
    body = re_commitify.sub(r"{{GitCommit|\1}}", body)

    return body


def gen_commit_pretty(c, unreported=None, rstate=None):
    body = gen_commit_message_pretty(c, unreported)

    if rstate is not None:
        return "* [%s] %s ({{GitCommit|rB%s}})." % (rstate, body, c.sha1.decode()[:10])
    return "* %s ({{GitCommit|rB%s}})." % (rstate, body, c.sha1.decode()[:10])


def gen_commit_unprettify(body):
    if body.startswith("* ["):
        end = body.find("]")
        if end > 0:
            body = body[end + 2:]  # +2 to remove ] itself, and following space.
    start = body.rfind("({{GitCommit|rB")
    if start > 0:
        body = body[:start - 1]  # -1 to remove trailing space.
    return body


def print_categories_tree():
    for i, (main_cat, sub_cats) in enumerate(BUGFIX_CATEGORIES):
        print("\t[%d] %s" % (i, main_cat))
        for j, sub_cat in enumerate(sub_cats):
            print("\t\t[%d] %s" % (j, sub_cat))


def release_log_extract_messages(path):
    messages = set()

    if os.path.exists(path):
        with open(path, 'r') as f:
            ignore = False
            header = True
            for l in f:
                if IGNORE_END_LINE in l:
                    ignore = False
                    continue
                elif ignore or IGNORE_START_LINE in l:
                    ignore = True
                    continue
                l = l.strip(" \n")
                if header and not l.startswith("=="):
                    continue  # Header, we don't care here.
                header = False
                if not l.startswith("==") and "Fix " in l:
                    messages.add(gen_commit_unprettify(l))

    return messages


def release_log_init(path, source_dir, blender_rev, start_sha1, end_sha1, rstate, rstate_list):
    from git_log import GitRepo

    if rstate is not None:
        header = "= Blender %s: Bug Fixes =\n\n" \
                 "[%s] Changes from revision {{GitCommit|rB%s}} to {{GitCommit|rB%s}}, inclusive.\n\n" \
                 % (blender_rev, rstate, start_sha1[:10], end_sha1[:10])
    else:
        header = "= Blender %s: Bug Fixes =\n\n" \
                 "Changes from revision {{GitCommit|rB%s}} to {{GitCommit|rB%s}}, inclusive.\n\n" \
                 % (blender_rev, start_sha1[:10], end_sha1[:10])

    release_log = {"__HEADER__": header, "__COUNT__": [0, 0], "__RSTATES__": {k: [] for k in rstate_list}}

    if os.path.exists(path):
        branch = GitRepo(source_dir).branch.decode().strip()

        sub_cats_to_main_cats = {s_cat: m_cat[0] for m_cat in BUGFIX_CATEGORIES for s_cat in m_cat[1]}
        main_cats = {m_cat[0] for m_cat in BUGFIX_CATEGORIES}
        with open(path, 'r') as f:
            header = []
            main_cat = None
            sub_cat = None
            ignore = False
            for l in f:
                if IGNORE_END_LINE in l:
                    ignore = False
                    continue
                elif ignore or IGNORE_START_LINE in l:
                    ignore = True
                    continue
                l = l.strip(" \n")
                if not header:
                    header.append(l)
                    for hl in f:
                        if IGNORE_END_LINE in hl:
                            ignore = False
                            continue
                        elif ignore or IGNORE_START_LINE in hl:
                            ignore = True
                            continue
                        hl = hl.strip(" \n")
                        if hl.startswith("=="):
                            main_cat = hl.strip(" =")
                            if main_cat not in main_cats:
                                sub_cat = main_cat
                                main_cat = sub_cats_to_main_cats.get(main_cat, None)
                            else:
                                sub_cat = None
                            #~ print("hl MAINCAT:", hl, main_cat, " | ", sub_cat)
                            break
                        header.append(hl)

                    if rstate is not None:
                        release_log["__HEADER__"] = "%s[%s] Changes from revision {{GitCommit|%s}} to " \
                                                    "{{GitCommit|%s}}, inclusive (''%s'' branch).\n\n" \
                                                    "" % ("\n".join(header), rstate,
                                                          start_sha1[:10], end_sha1[:10], branch)
                    else:
                        release_log["__HEADER__"] = "%sChanges from revision {{GitCommit|%s}} to {{GitCommit|%s}}, " \
                                                    "inclusive (''%s'' branch).\n\n" \
                                                    "" % ("\n".join(header), start_sha1[:10], end_sha1[:10], branch)
                    count = release_log["__COUNT__"] = [0, 0]
                    continue

                if l.startswith("==="):
                    sub_cat = l.strip(" =")
                    if sub_cat in sub_cats_to_main_cats:
                        main_cat = sub_cats_to_main_cats.get(sub_cat, None)
                    elif sub_cat in main_cats:
                        main_cat = sub_cat
                        sub_cat = None
                    else:
                        main_cat = None
                    #~ print("l SUBCAT:", l, main_cat, " | ", sub_cat)
                elif l.startswith("=="):
                    main_cat = l.strip(" =")
                    if main_cat not in main_cats:
                        sub_cat = main_cat
                        main_cat = sub_cats_to_main_cats.get(main_cat, None)
                    else:
                        sub_cat = None
                    #~ print("l MAINCAT:", l, main_cat, " | ", sub_cat)
                elif "Fix " in l:
                    if "Fix {{BugReport|" in l:
                        main_cat_data, _ = release_log.setdefault(main_cat, ({}, {}))
                        main_cat_data.setdefault(sub_cat, []).append(l)
                        count[0] += 1
                        #~ print("l REPORTED:", l)
                    else:
                        _, main_cat_data_unreported = release_log.setdefault(main_cat, ({}, {}))
                        main_cat_data_unreported.setdefault(sub_cat, []).append(l)
                        count[1] += 1
                        #~ print("l UNREPORTED:", l)
                    l_rstate = l.strip("* ")
                    if l_rstate.startswith("["):
                        end = l_rstate.find("]")
                        if end > 0:
                            rstate = l_rstate[1:end]
                            if rstate in release_log["__RSTATES__"]:
                                release_log["__RSTATES__"][rstate].append("* %s" % l_rstate[end + 1:].strip())

    return release_log


def write_release_log(path, release_log, c, cat, rstate, rstate_list):
    import io

    main_cat, sub_cats = BUGFIX_CATEGORIES[cat[0]]
    sub_cat = sub_cats[cat[1]] if cat[1] is not None else None

    main_cat_data, main_cat_data_unreported = release_log.setdefault(main_cat, ({}, {}))
    unreported = [False]
    entry = gen_commit_pretty(c, unreported, rstate)
    if unreported[0]:
        main_cat_data_unreported.setdefault(sub_cat, []).append(entry)
        release_log["__COUNT__"][1] += 1
    else:
        main_cat_data.setdefault(sub_cat, []).append(entry)
        release_log["__COUNT__"][0] += 1

    if rstate in release_log["__RSTATES__"]:
        release_log["__RSTATES__"][rstate].append(gen_commit_pretty(c))

    lines = []
    main_cat_lines = []
    sub_cat_lines = []
    for main_cat, sub_cats in BUGFIX_CATEGORIES:
        main_cat_data = release_log.get(main_cat, ({}, {}))
        main_cat_lines[:] = ["== %s ==" % main_cat]
        for data in main_cat_data:
            entries = data.get(None, [])
            if entries:
                main_cat_lines.extend(entries)
                main_cat_lines.append("")
        if len(main_cat_lines) == 1:
            main_cat_lines.append("")
        for sub_cat in sub_cats:
            sub_cat_lines[:] = ["=== %s ===" % sub_cat]
            for data in main_cat_data:
                entries = data.get(sub_cat, [])
                if entries:
                    sub_cat_lines.extend(entries)
                    sub_cat_lines.append("")
            if len(sub_cat_lines) > 2:
                main_cat_lines += sub_cat_lines
        if len(main_cat_lines) > 2:
            lines += main_cat_lines

    if None in release_log:
        main_cat_data = release_log.get(None, ({}, {}))
        main_cat_lines[:] = ["== %s ==\n\n" % "UNSORTED"]
        for data in main_cat_data:
            entries = data.get(None, [])
            if entries:
                main_cat_lines.extend(entries)
                main_cat_lines.append("")
        if len(main_cat_lines) > 2:
            lines += main_cat_lines

    with open(path, 'w') as f:
        f.write(release_log["__HEADER__"])

        count = release_log["__COUNT__"]
        f.write("%s\n" % IGNORE_START_LINE)
        f.write("Total fixed bugs: %d (%d from tracker, %d reported/found by other ways).\n\n"
                "" % (sum(count), count[0], count[1]))
        f.write("%s\n%s\n\n" % ("{{Note|Note|Before RC1 (i.e. during regular development of next version in master "
                                "branch), only fixes of issues which already existed in previous official releases are "
                                "listed here. Fixes for regressions introduced since last release, or for new "
                                "features, are '''not''' listed here.<br/>For following RCs and final release, "
                                "'''all''' backported fixes are listed.}}", IGNORE_END_LINE))

        f.write("\n".join(lines))
        f.write("\n")

        f.write("%s\n\n<hr/>\n\n" % IGNORE_START_LINE)
        for rst in rstate_list:
            entries = release_log["__RSTATES__"].get(rst, [])
            if entries:
                f.write("== %s ==\n" % rst)
                f.write("For %s, %d bugs were fixed:\n\n" % (rst, len(entries)))
                f.write("\n".join(entries))
                f.write("\n\n")
        f.write("%s\n" % IGNORE_END_LINE)


def argparse_create():
    import argparse
    global __doc__

    # When --help or no args are given, print this help
    usage_text = __doc__

    epilog = "This script is typically used to help write release notes"

    parser = argparse.ArgumentParser(description=usage_text, epilog=epilog,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument(
        "--source", dest="source_dir",
        metavar='PATH', required=True,
        help="Path to git repository")
    parser.add_argument(
        "--range", dest="range_sha1",
                        metavar='SHA1_RANGE', required=False,
                        help="Range to use, eg: 169c95b8..HEAD")
    parser.add_argument(
        "--author", dest="author",
        metavar='AUTHOR', type=str, required=False,
        help=("Author(s) to filter commits ("))
    parser.add_argument(
        "--filter", dest="filter_type",
        metavar='FILTER', type=str, required=False,
        help=("Method to filter commits in ['BUGFIX', 'NOISE']"))
    parser.add_argument(
        "--filter-exclude-sha1", dest="filter_exclude_sha1_list",
        default=[], required=False, type=lambda s: s.split(","),
        help=("Coma-separated list of commits to ignore/skip"))
    parser.add_argument(
        "--filter-exclude-sha1-fromfiles", dest="filter_exclude_sha1_filepaths",
        default="", required=False, nargs='*',
        help=("One or more text files storing list of commits to ignore/skip"))
    parser.add_argument(
        "--filter-exclude-fromreleaselogs", dest="filter_exclude_releaselogs",
        default="", required=False, nargs='*',
        help=("One or more text files storing release logs, to ignore/skip their entries "
              "(based on message comparison, not commit sha1)"))
    parser.add_argument(
        "--accept-log", dest="accept_log",
        default=False, action='store_true', required=False,
        help=("Also output more complete info about accepted commits (summary, author...)"))
    parser.add_argument(
        "--accept-pretty", dest="accept_pretty",
        default=False, action='store_true', required=False,
        help=("Also output pretty-printed accepted commits (nearly ready for WIKI release notes)"))
    parser.add_argument(
        "--accept-releaselog", dest="accept_releaselog",
        default=False, action='store_true', required=False,
        help=("Also output accepted commits as a wiki release log page (adds sorting by categories)"))
    parser.add_argument(
        "--blender-rev", dest="blender_rev",
        default=None, required=False,
        help=("Blender revision (only used to generate release notes page)"))
    parser.add_argument(
        "--blender-rstate", dest="blender_rstate",
        default="alpha", required=False,
        help=("Blender release state (like alpha, beta, rc1, final, corr_a, corr_b, etc.), "
              "each revision will be tagged by given one"))
    parser.add_argument(
        "--blender-rstate-list", dest="blender_rstate_list",
        default="", required=False, type=lambda s: s.split(","),
        help=("Blender release state(s) to additionally list in their own sections "
              "(e.g. pass 'RC2' to list fixes between RC1 and RC2, ie tagged as RC2, etc.)"))

    return parser


def main():
    # ----------
    # Parse Args

    args = argparse_create().parse_args()

    for path in args.filter_exclude_sha1_filepaths:
        if os.path.exists(path):
            with open(path, 'r') as f:
                args.filter_exclude_sha1_list += [sha1 for l in f for sha1 in l.split()]
    args.filter_exclude_sha1_list = {sha1.encode() for sha1 in args.filter_exclude_sha1_list}

    messages = set()
    for path in args.filter_exclude_releaselogs:
        messages |= release_log_extract_messages(path)
    args.filter_exclude_releaselogs = messages

    from git_log import GitCommit, GitCommitIter

    # --------------
    # Filter Commits

    def match(c):
        # filter_type
        if not args.filter_type:
            pass
        elif args.filter_type == 'BUGFIX':
            first_line = c.body.split("\n\n")[0].strip(" :.;-\n").replace("\n", " ")
            assert(len(first_line))
            if any(w for w in first_line.split() if w.lower().startswith(("fix", "bugfix", "bug-fix"))):
                pass
            else:
                return False
        elif args.filter_type == 'NOISE':
            first_line = c.body.strip().split("\n")[0]
            assert(len(first_line))
            if any(w for w in first_line.split() if w.lower().startswith("cleanup")):
                pass
            else:
                return False
        else:
            raise Exception("Filter type %r isn't known" % args.filter_type)

        # author
        if not args.author:
            pass
        elif args.author != c.author:
            return False

        # commits to exclude
        if c.sha1 in args.filter_exclude_sha1_list:
            return False

        # exclude by commit message (because cherry-pick totally breaks relations with original commit...)
        if args.filter_exclude_releaselogs:
            if gen_commit_message_pretty(c) in args.filter_exclude_releaselogs:
                return False

        return True

    if args.accept_releaselog:
        blender_rev = args.blender_rev or "<UNKNOWN>"
        commits = tuple(GitCommitIter(args.source_dir, args.range_sha1))
        release_log = release_log_init(ACCEPT_RELEASELOG_FILE, args.source_dir, blender_rev,
                                       commits[-1].sha1.decode(), commits[0].sha1.decode(),
                                       args.blender_rstate, args.blender_rstate_list)
        commits = [c for c in commits if match(c)]
    else:
        commits = [c for c in GitCommitIter(args.source_dir, args.range_sha1) if match(c)]

    # oldest first
    commits.reverse()

    tot_accept = 0
    tot_reject = 0

    def exit_message():
        print("  Written",
              colorize(ACCEPT_FILE, color='green'), "(%d)" % tot_accept,
              colorize(ACCEPT_LOG_FILE, color='yellow'), "(%d)" % tot_accept,
              colorize(ACCEPT_PRETTY_FILE, color='blue'), "(%d)" % tot_accept,
              colorize(REJECT_FILE, color='red'), "(%d)" % tot_reject,
              )

    def get_cat(ch, max_idx):
        cat = -1
        try:
            cat = int(ch)
        except:
            pass
        if 0 <= cat < max_idx:
            return cat
        print("Invalid input %r" % ch)
        return None

    for i, c in enumerate(commits):
        if os.name == "posix":
            # also clears scrollback
            os.system("tput reset")
        else:
            print('\x1b[2J')  # clear

        sha1 = c.sha1

        # diff may scroll off the screen, that's OK
        os.system("git --git-dir %s show %s --format=%%n" % (c._git_dir, sha1.decode('ascii')))
        print("")
        print_commit(c)
        sys.stdout.flush()

        accept = False
        while True:
            print("Space=" + colorize("Accept", 'green'),
                  "Enter=" + colorize("Skip", 'red'),
                  "Ctrl+C or X=" + colorize("Exit", color='white'),
                  "[%d of %d]" % (i + 1, len(commits)),
                  "(+%d | -%d)" % (tot_accept, tot_reject),
                  )
            ch = getch()

            if ch == b'\x03' or ch == b'x':
                # Ctrl+C
                exit_message()
                print("Goodbye! (%s)" % c.sha1.decode())
                return False
            elif ch == b' ':
                log_filepath = ACCEPT_FILE
                log_filepath_log = ACCEPT_LOG_FILE
                log_filepath_pretty = ACCEPT_PRETTY_FILE
                tot_accept += 1

                if args.accept_releaselog:  # Enter sub-loop for category selection.
                    done_main = True
                    c1 = c2 = None
                    while True:
                        if c1 is None:
                            print("Select main category (V=View all categories, M=Commit message): \n\t%s"
                                  "" % " | ".join("[%d] %s" % (i, cat[0]) for i, cat in enumerate(BUGFIX_CATEGORIES)))
                        else:
                            main_cat = BUGFIX_CATEGORIES[c1][0]
                            sub_cats = BUGFIX_CATEGORIES[c1][1]
                            if not sub_cats:
                                break
                            print("[%d] %s: Select sub category "
                                  "(V=View all categories, M=Commit message, Enter=No sub-categories, "
                                  "Backspace=Select other main category): \n\t%s"
                                  "" % (c1, main_cat,
                                        " | ".join("[%d] %s" % (i, cat) for i, cat in enumerate(sub_cats))))

                        ch = getch()

                        if ch == b'\x7f':  # backspace
                            done_main = False
                            break
                        elif ch == b'\x03' or ch == b'x':
                            # Ctrl+C
                            exit_message()
                            print("Goodbye! (%s)" % c.sha1.decode())
                            return
                        elif ch == b'v':
                            print_categories_tree()
                            print("")
                        elif ch == b'm':
                            print_commit(c)
                            print("")
                        elif c1 is None:
                            c1 = get_cat(ch, len(BUGFIX_CATEGORIES))
                        elif c2 is None:
                            if ch == b'\r':
                                break
                            elif ch == b'\x7f':  # backspace
                                c1 = None
                                continue
                            c2 = get_cat(ch, len(BUGFIX_CATEGORIES[c1][1]))
                            if c2 is not None:
                                break
                        else:
                            print("BUG! this should not happen!")

                    if done_main is False:
                        # Go back to main loop, this commit is no more accepted nor rejected.
                        tot_accept -= 1
                        continue

                    write_release_log(ACCEPT_RELEASELOG_FILE, release_log, c, (c1, c2),
                                      args.blender_rstate, args.blender_rstate_list)
                break
            elif ch == b'\r':
                log_filepath = REJECT_FILE
                log_filepath_log = None
                log_filepath_pretty = None
                tot_reject += 1
                break
            else:
                print("Invalid input %r" % ch)

        with open(log_filepath, 'ab') as f:
            f.write(sha1 + b'\n')

        if args.accept_pretty and log_filepath_pretty:
            with open(log_filepath_pretty, 'a') as f:
                f.write(gen_commit_pretty(c, rstate=args.blender_rstate) + "\n")

        if args.accept_log and log_filepath_log:
            with open(log_filepath_log, 'a') as f:
                f.write(gen_commit_log(c) + "\n")

    exit_message()


if __name__ == "__main__":
    main()



# GPL # Original Authors: Evan J. Rosky (syrux), Chichiri, Jace Priester #

import bpy
import random
import math
from bpy.types import (
                Operator,
                Menu,
                )
from mathutils import (
                Vector,
                Quaternion,
                )
from bpy.props import (
                BoolProperty,
                IntProperty,
                FloatProperty,
                StringProperty,
                )

# ################### Globals #################### #

doprots = True

# Datas in which we will build the new discombobulated mesh
nPolygons = []
nVerts = []
Verts = []
Polygons = []
dVerts = []
dPolygons = []
i_prots = []        # index of the top polygons on which we"ll generate the doodads
i_dood_type = []    # type of doodad (given by index of the doodad obj)


# ############### Utility Functions ############### #

def randnum(a, b):
    return random.random() * (b - a) + a


def randVertex(a, b, c, d, Verts):
    """Return a vector of a random vertex on a quad-polygon"""
    i = random.randint(1, 2)
    A, B, C, D = 0, 0, 0, 0
    if(a == 1):
        A, B, C, D = a, b, c, d
    else:
        A, B, C, D = a, d, c, b

    i = randnum(0.1, 0.9)

    vecAB = Verts[B] - Verts[A]
    E = Verts[A] + vecAB * i

    vecDC = Verts[C] - Verts[D]
    F = Verts[D] + vecDC * i

    i = randnum(0.1, 0.9)
    vecEF = F - E

    O = E + vecEF * i
    return O


# ################## Protusions #################### #

def fill_older_datas(verts, polygon):
    """ Specifically coded to be called by the function addProtusionToPolygon,
        its sets up a tuple which contains the vertices from the base and the top of the protusions.
    """
    temp_vertices = []
    temp_vertices.append(verts[polygon[0]].copy())
    temp_vertices.append(verts[polygon[1]].copy())
    temp_vertices.append(verts[polygon[2]].copy())
    temp_vertices.append(verts[polygon[3]].copy())
    temp_vertices.append(verts[polygon[0]].copy())
    temp_vertices.append(verts[polygon[1]].copy())
    temp_vertices.append(verts[polygon[2]].copy())
    temp_vertices.append(verts[polygon[3]].copy())
    return temp_vertices


def extrude_top(temp_vertices, normal, height):
    """ This function extrude the polygon composed of the four first members of the tuple
        temp_vertices along the normal multiplied by the height of the extrusion.
    """
    j = 0
    while j < 3:
        temp_vertices[0][j] += normal[j] * height
        temp_vertices[1][j] += normal[j] * height
        temp_vertices[2][j] += normal[j] * height
        temp_vertices[3][j] += normal[j] * height
        j += 1


def scale_top(temp_vertices, center, normal, height, scale_ratio):
    """ This function scale the polygon composed of the four first members of the tuple temp_vertices. """
    vec1 = [0, 0, 0]
    vec2 = [0, 0, 0]
    vec3 = [0, 0, 0]
    vec4 = [0, 0, 0]

    j = 0
    while j < 3:
        center[j] += normal[j] * height
        vec1[j] = temp_vertices[0][j] - center[j]
        vec2[j] = temp_vertices[1][j] - center[j]
        vec3[j] = temp_vertices[2][j] - center[j]
        vec4[j] = temp_vertices[3][j] - center[j]
        temp_vertices[0][j] = center[j] + vec1[j] * (1 - scale_ratio)
        temp_vertices[1][j] = center[j] + vec2[j] * (1 - scale_ratio)
        temp_vertices[2][j] = center[j] + vec3[j] * (1 - scale_ratio)
        temp_vertices[3][j] = center[j] + vec4[j] * (1 - scale_ratio)
        j += 1


def add_prot_polygons(temp_vertices):
    """ Specifically coded to be called by addProtusionToPolygon, this function
        put the data from the generated protusion at the end the tuples Verts and Polygons,
        which will later used to generate the final mesh.
    """
    global Verts
    global Polygons
    global i_prots

    findex = len(Verts)
    Verts += temp_vertices

    polygontop = [findex + 0, findex + 1, findex + 2, findex + 3]
    polygon1 = [findex + 0, findex + 1, findex + 5, findex + 4]
    polygon2 = [findex + 1, findex + 2, findex + 6, findex + 5]
    polygon3 = [findex + 2, findex + 3, findex + 7, findex + 6]
    polygon4 = [findex + 3, findex + 0, findex + 4, findex + 7]

    Polygons.append(polygontop)
    i_prots.append(len(Polygons) - 1)
    Polygons.append(polygon1)
    Polygons.append(polygon2)
    Polygons.append(polygon3)
    Polygons.append(polygon4)


def addProtusionToPolygon(obpolygon, verts, minHeight, maxHeight, minTaper, maxTaper):
    """Create a protusion from the polygon "obpolygon" of the original object and use
       several values sent by the user. It calls in this order the following functions:
       - fill_older_data;
       - extrude_top;
       - scale_top;
       - add_prot_polygons;
   """
    # some useful variables
    polygon = obpolygon.vertices

    tVerts = list(fill_older_datas(verts, polygon))     # list of temp vertices
    height = randnum(minHeight, maxHeight)              # height of generated protusion
    scale_ratio = randnum(minTaper, maxTaper)

    # extrude the top polygon
    extrude_top(tVerts, obpolygon.normal, height)
    # Now, we scale, the top polygon along its normal
    scale_top(tVerts, GetPolyCentroid(obpolygon, verts), obpolygon.normal, height, scale_ratio)
    # Finally, we add the protusions to the list of polygons
    add_prot_polygons(tVerts)


# ################# Divide a polygon ############### #

def divide_one(list_polygons, list_vertices, verts, polygon, findex):
    """ called by divide_polygon, to generate a polygon from one polygon, maybe I could simplify this process """
    temp_vertices = []
    temp_vertices.append(verts[polygon[0]].copy())
    temp_vertices.append(verts[polygon[1]].copy())
    temp_vertices.append(verts[polygon[2]].copy())
    temp_vertices.append(verts[polygon[3]].copy())

    list_vertices += temp_vertices

    list_polygons.append([findex + 0, findex + 1, findex + 2, findex + 3])


def divide_two(list_polygons, list_vertices, verts, polygon, findex):
    """ called by divide_polygon, to generate two polygons from one polygon and
        add them to the list of polygons and vertices which form the discombobulated mesh
    """
    temp_vertices = []
    temp_vertices.append(verts[polygon[0]].copy())
    temp_vertices.append(verts[polygon[1]].copy())
    temp_vertices.append(verts[polygon[2]].copy())
    temp_vertices.append(verts[polygon[3]].copy())
    temp_vertices.append((verts[polygon[0]] + verts[polygon[1]]) / 2)
    temp_vertices.append((verts[polygon[2]] + verts[polygon[3]]) / 2)

    list_vertices += temp_vertices

    list_polygons.append([findex + 0, findex + 4, findex + 5, findex + 3])
    list_polygons.append([findex + 1, findex + 2, findex + 5, findex + 4])


def divide_three(list_polygons, list_vertices, verts, polygon, findex, center):
    """ called by divide_polygon, to generate three polygons from one polygon and
        add them to the list of polygons and vertices which form the discombobulated mesh
    """
    temp_vertices = []
    temp_vertices.append(verts[polygon[0]].copy())
    temp_vertices.append(verts[polygon[1]].copy())
    temp_vertices.append(verts[polygon[2]].copy())
    temp_vertices.append(verts[polygon[3]].copy())
    temp_vertices.append((verts[polygon[0]] + verts[polygon[1]]) / 2)
    temp_vertices.append((verts[polygon[2]] + verts[polygon[3]]) / 2)
    temp_vertices.append((verts[polygon[1]] + verts[polygon[2]]) / 2)
    temp_vertices.append(center.copy())

    list_vertices += temp_vertices

    list_polygons.append([findex + 0, findex + 4, findex + 5, findex + 3])
    list_polygons.append([findex + 1, findex + 6, findex + 7, findex + 4])
    list_polygons.append([findex + 6, findex + 2, findex + 5, findex + 7])


def divide_four(list_polygons, list_vertices, verts, polygon, findex, center):
    """ called by divide_polygon, to generate four polygons from one polygon and
        add them to the list of polygons and vertices which form the discombobulated mesh
    """
    temp_vertices = []
    temp_vertices.append(verts[polygon[0]].copy())
    temp_vertices.append(verts[polygon[1]].copy())
    temp_vertices.append(verts[polygon[2]].copy())
    temp_vertices.append(verts[polygon[3]].copy())
    temp_vertices.append((verts[polygon[0]] + verts[polygon[1]]) / 2)
    temp_vertices.append((verts[polygon[2]] + verts[polygon[3]]) / 2)
    temp_vertices.append((verts[polygon[1]] + verts[polygon[2]]) / 2)
    temp_vertices.append(center.copy())
    temp_vertices.append((verts[polygon[0]] + verts[polygon[3]]) / 2)
    temp_vertices.append(center.copy())

    list_vertices += temp_vertices

    list_polygons.append([findex + 0, findex + 4, findex + 7, findex + 8])
    list_polygons.append([findex + 1, findex + 6, findex + 7, findex + 4])
    list_polygons.append([findex + 6, findex + 2, findex + 5, findex + 7])
    list_polygons.append([findex + 8, findex + 7, findex + 5, findex + 3])


def dividepolygon(obpolygon, verts, number):
    """Divide the poly into the wanted number of polygons"""
    global nPolygons
    global nVerts

    poly = obpolygon.vertices

    if(number == 1):
        divide_one(nPolygons, nVerts, verts, poly, len(nVerts))
    elif(number == 2):
        divide_two(nPolygons, nVerts, verts, poly, len(nVerts))
    elif(number == 3):
        divide_three(nPolygons, nVerts, verts, poly, len(nVerts), GetPolyCentroid(obpolygon, verts))
    elif(number == 4):
        divide_four(nPolygons, nVerts, verts, poly, len(nVerts), GetPolyCentroid(obpolygon, verts))


# ################## Discombobulate ################ #

def GetPolyCentroid(obpolygon, allvertcoords):
    centroid = Vector((0, 0, 0))
    for vindex in obpolygon.vertices:
        centroid += Vector(allvertcoords[vindex])
    centroid /= len(obpolygon.vertices)
    return centroid


def division(obpolygons, verts, sf1, sf2, sf3, sf4):
    """Function to divide each of the selected polygons"""
    divide = []
    if (sf1):
        divide.append(1)
    if (sf2):
        divide.append(2)
    if (sf3):
        divide.append(3)
    if (sf4):
        divide.append(4)

    for poly in obpolygons:
        if(poly.select is True and len(poly.vertices) == 4):
            a = random.randint(0, len(divide) - 1)
            dividepolygon(poly, verts, divide[a])


def protusion(obverts, obpolygons, minHeight, maxHeight, minTaper, maxTaper):
    """function to generate the protusions"""
    verts = []
    for vertex in obverts:
        verts.append(vertex.co)

    for polygon in obpolygons:
        if(polygon.select is True):
            if(len(polygon.vertices) == 4):
                addProtusionToPolygon(polygon, verts, minHeight, maxHeight, minTaper, maxTaper)


def test_v2_near_v1(v1, v2):
    if (v1.x - 0.1 <= v2.x <= v1.x + 0.1 and
       v1.y - 0.1 <= v2.y <= v1.y + 0.1 and
       v1.z - 0.1 <= v2.z <= v1.z + 0.1):
        return True

    return False


def angle_between_nor(nor_orig, nor_result):
    angle = math.acos(nor_orig.dot(nor_result))
    axis = nor_orig.cross(nor_result).normalized()

    q = Quaternion()
    q.x = axis.x * math.sin(angle / 2)
    q.y = axis.y * math.sin(angle / 2)
    q.z = axis.z * math.sin(angle / 2)
    q.w = math.cos(angle / 2)

    return q


def doodads(self, object1, mesh1, dmin, dmax):
    """function to generate the doodads"""
    global dVerts
    global dPolygons
    i = 0
    # on parcoure cette boucle pour ajouter des doodads a toutes les polygons
    # english translation: this loops adds doodads to all polygons
    while(i < len(object1.data.polygons)):
        if object1.data.polygons[i].select is False:
            continue

        doods_nbr = random.randint(dmin, dmax)
        j = 0

        while(j <= doods_nbr):
            origin_dood = randVertex(object1.data.polygons[i].vertices[0], object1.data.polygons[i].vertices[1],
                                     object1.data.polygons[i].vertices[2], object1.data.polygons[i].vertices[3], Verts)
            type_dood = random.randint(0, len(self.DISC_doodads) - 1)
            polygons_add = []
            verts_add = []

            # First we have to apply scaling and rotation to the mesh
            bpy.ops.object.select_pattern(pattern=self.DISC_doodads[type_dood], extend=False)
            bpy.context.view_layer.objects.active = bpy.data.objects[self.DISC_doodads[type_dood]]
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

            for polygon in bpy.data.objects[self.DISC_doodads[type_dood]].data.polygons:
                polygons_add.append(polygon.vertices)
            for vertex in bpy.data.objects[self.DISC_doodads[type_dood]].data.vertices:
                verts_add.append(vertex.co.copy())
            normal_original_polygon = object1.data.polygons[i].normal

            nor_def = Vector((0.0, 0.0, 1.0))
            qr = nor_def.rotation_difference(normal_original_polygon.normalized())

            if(test_v2_near_v1(nor_def, -normal_original_polygon)):
                qr = Quaternion((0.0, 0.0, 0.0, 0.0))

            # qr = angle_between_nor(nor_def, normal_original_polygon)
            for vertex in verts_add:
                vertex.rotate(qr)
                vertex += origin_dood
            findex = len(dVerts)

            for polygon in polygons_add:
                dPolygons.append([polygon[0] + findex, polygon[1] + findex, polygon[2] + findex, polygon[3] + findex])
                i_dood_type.append(bpy.data.objects[self.DISC_doodads[type_dood]].name)

            for vertex in verts_add:
                dVerts.append(vertex)
            j += 1
        i += 5


def protusions_repeat(object1, mesh1, r_prot):

        for j in i_prots:
            if j < len(object1.data.polygons):
                object1.data.polygons[j].select = True
            else:
                print("Warning: hit end of polygons in object1")


# add material to discombobulated mesh
def setMatProt(discObj, origObj, sideProtMat, topProtMat):
    # First we put the materials in their slots
    bpy.ops.object.select_pattern(pattern=discObj.name, extend=False)
    bpy.context.view_layer.objects.active = bpy.data.objects[discObj.name]
    try:
        origObj.material_slots[topProtMat]
        origObj.material_slots[sideProtMat]
    except:
        return

    bpy.ops.object.material_slot_add()
    bpy.ops.object.material_slot_add()
    discObj.material_slots[0].material = origObj.material_slots[topProtMat].material
    discObj.material_slots[1].material = origObj.material_slots[sideProtMat].material

    # Then we assign materials to protusions
    for polygon in discObj.data.polygons:
        if polygon.index in i_prots:
            polygon.material_index = 0
        else:
            polygon.material_index = 1


def setMatDood(self, doodObj):
    # First we add the materials slots
    bpy.ops.object.select_pattern(pattern=doodObj.name, extend=False)
    bpy.context.view_layer.objects.active = doodObj
    for name in self.DISC_doodads:
        try:
            bpy.ops.object.material_slot_add()
            doodObj.material_slots[-1].material = bpy.data.objects[name].material_slots[0].material
            for polygon in doodObj.data.polygons:
                if i_dood_type[polygon.index] == name:
                    polygon.material_index = len(doodObj.material_slots) - 1
        except:
            print()


def clean_doodads(self):
    current_doodads = list(self.DISC_doodads)

    for name in current_doodads:
        if name not in bpy.data.objects:
            self.DISC_doodads.remove(name)


def discombobulate(self, minHeight, maxHeight, minTaper, maxTaper, sf1, sf2, sf3, sf4,
                   dmin, dmax, r_prot, sideProtMat, topProtMat, isLast):
    global doprots
    global nVerts
    global nPolygons
    global Verts
    global Polygons
    global dVerts
    global dPolygons
    global i_prots

    bpy.ops.object.mode_set(mode="OBJECT")

    # start by cleaning up doodads that don"t exist anymore
    clean_doodads(self)

    # Create the discombobulated mesh
    mesh = bpy.data.meshes.new("tmp")
    object = bpy.data.objects.new("tmp", mesh)
    bpy.context.collection.objects.link(object)

    # init final verts and polygons tuple
    nPolygons = []
    nVerts = []
    Polygons = []
    Verts = []
    dPolygons = []
    dVerts = []

    origObj = bpy.context.active_object

    # There we collect the rotation, translation and scaling datas from the original mesh
    to_translate = bpy.context.active_object.location
    to_scale = bpy.context.active_object.scale
    to_rotate = bpy.context.active_object.rotation_euler

    # First, we collect all the information we will need from the previous mesh
    obverts = bpy.context.active_object.data.vertices
    obpolygons = bpy.context.active_object.data.polygons
    verts = []
    for vertex in obverts:
        verts.append(vertex.co)

    division(obpolygons, verts, sf1, sf2, sf3, sf4)

    # Fill in the discombobulated mesh with the new polygons
    mesh.from_pydata(nVerts, [], nPolygons)
    mesh.update(calc_edges=True)

    # Reload the datas
    bpy.ops.object.select_all(action="DESELECT")
    bpy.ops.object.select_pattern(pattern=object.name, extend=False)
    bpy.context.view_layer.objects.active = bpy.data.objects[object.name]
    obverts = bpy.context.active_object.data.vertices
    obpolygons = bpy.context.active_object.data.polygons

    protusion(obverts, obpolygons, minHeight, maxHeight, minTaper, maxTaper)

    # Fill in the discombobulated mesh with the new polygons
    mesh1 = bpy.data.meshes.new("discombobulated_object")
    object1 = bpy.data.objects.new("discombobulated_mesh", mesh1)
    bpy.context.collection.objects.link(object1)
    mesh1.from_pydata(Verts, [], Polygons)
    mesh1.update(calc_edges=True)

    # Set the material"s of discombobulated object
    setMatProt(object1, origObj, sideProtMat, topProtMat)

    bpy.ops.object.select_pattern(pattern=object1.name, extend=False)
    bpy.context.view_layer.objects.active = bpy.data.objects[object1.name]
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.normals_make_consistent(inside=False)
    bpy.ops.mesh.select_all(action="DESELECT")
    bpy.ops.object.mode_set(mode="OBJECT")

    # if(bpy.context.scene.repeatprot):
    protusions_repeat(object1, mesh1, r_prot)

    if(len(self.DISC_doodads) != 0 and self.dodoodads and isLast):
        doodads(self, object1, mesh1, dmin, dmax)
        mesh2 = bpy.data.meshes.new("dood_mesh")
        object2 = bpy.data.objects.new("dood_obj", mesh2)
        bpy.context.collection.objects.link(object2)
        mesh2.from_pydata(dVerts, [], dPolygons)
        mesh2.update(calc_edges=True)
        setMatDood(self, object2)
        object2.location = to_translate
        object2.rotation_euler = to_rotate
        object2.scale = to_scale

    bpy.ops.object.select_pattern(pattern=object.name, extend=False)
    bpy.context.view_layer.objects.active = bpy.data.objects[object.name]
    bpy.ops.object.delete()

    bpy.ops.object.select_pattern(pattern=object1.name, extend=False)
    bpy.context.view_layer.objects.active = bpy.data.objects[object1.name]
    bpy.context.view_layer.update()

    # translate, scale and rotate discombobulated results
    object1.location = to_translate
    object1.rotation_euler = to_rotate
    object1.scale = to_scale

    # set all polys to selected. this allows recursive discombobulating.
    for poly in mesh1.polygons:
        poly.select = True


# ### Operators for selecting and deselecting an object as a doodad ### #

class chooseDoodad(Operator):
    bl_idname = "object.discombobulate_set_doodad"
    bl_label = "Discombobulate set doodad object"
    bl_description = ("Save the Active Object as Doodad \n"
                      "Object has to be quads only")
    bl_options = {"REGISTER"}

    @classmethod
    def poll(cls, context):
        obj = bpy.context.active_object
        if (obj is not None and obj.type == "MESH"):
            mesh = obj.data

            for polygon in mesh.polygons:
                is_ok = len(polygon.vertices)
                if is_ok != 4:
                    return False
            return True

        return False

    def execute(self, context):
        obj_name = bpy.context.active_object.name
        msg = "Object with this name already saved"

        DISC_doodads = context.scene.discombobulator.DISC_doodads

        if obj_name not in DISC_doodads:
            DISC_doodads.append(obj_name)
            msg = "Saved Doodad object: {}".format(obj_name)

        self.report({"INFO"}, message=msg)

    def invoke(self, context, event):
        self.execute(context)
        return {"FINISHED"}


class unchooseDoodad(Operator):
    bl_idname = "object.discombobulate_unset_doodad"
    bl_label = "Discombobulate unset doodad object"
    bl_description = "Remove the saved Doodad Object(s)"
    bl_options = {"REGISTER"}

    remove_all: bpy.props.BoolProperty(
                        name="Remove all Doodads",
                        default=False,
                        )

    def execute(self, context):
        msg = ("No doodads to remove")
        DISC_doodads = context.scene.discombobulator.DISC_doodads
        if len(DISC_doodads) > 0:
            if not self.remove_all:
                name = bpy.context.active_object.name
                if name in DISC_doodads:
                    DISC_doodads.remove(name)
                    msg = ("Removed Doodad object: {}".format(name))
            else:
                DISC_doodads[:] = []
                msg = "Removed all Doodads"
        else:
            msg = "No Doodads to Remove"

        self.report({"INFO"}, message=msg)

    def invoke(self, context, event):
        self.execute(context)
        return {"FINISHED"}


# ################## Interpolygon ################## #

class discombobulator(Operator):
    bl_idname = "object.discombobulate"
    bl_label = "Discombobulate"
    bl_description = "Apply"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        i = 0
        while i < self.repeatprot:
            isLast = False
            if i == self.repeatprot - 1:
                isLast = True
            discombobulate(self.minHeight, self.maxHeight, self.minTaper, self.maxTaper, self.subpolygon1,
                           self.subpolygon2, self.subpolygon3, self.subpolygon4, self.mindoodads, self.maxdoodads,
                           self.repeatprot, self.sideProtMat, self.topProtMat, isLast)
            i += 1
        return {"FINISHED"}


class discombobulator_dodads_list(Menu):
    bl_idname = "OBJECT_MT_discombobulator_dodad_list"
    bl_label = "List of saved Doodads"
    bl_description = "List of the saved Doodad Object Names"
    bl_options = {"REGISTER"}

    def draw(self, context):
        layout = self.layout

        DISC_doodads = context.scene.discombobulator.DISC_doodads

        doodle = len(DISC_doodads)
        layout.label(text="Saved doodads : {}".format(doodle))
        layout.separator()
        if doodle > 0:
            for name in DISC_doodads:
                layout.label(text=name)


class discombob_help(Menu):
    bl_idname = "HELP_MT_discombobulator"
    bl_label = "Usage Information"
    bl_description = "Help"
    bl_options = {"REGISTER"}

    def draw(self, context):
        layout = self.layout
        layout.label(text="Usage Information:", icon="INFO")
        layout.separator()
        layout.label(text="Quads only, not Triangles or Ngons", icon="ERROR")
        layout.label(text="Works only with Mesh object that have faces")
        layout.separator()
        layout.label(text="Select a face or faces")
        layout.label(text="Press Discombobulate to create greebles")
        layout.label(text="In object mode, still needs a selection in Edit Mode")
        layout.separator()
        layout.label(text="Doodads - additional objects layered on the mesh surface")
        layout.label(text="(Similar to dupliverts - but as one separate object)")
        layout.separator()
        layout.label(text="Limitations:", icon="MOD_EXPLODE")
        layout.label(text="Be careful with the repeat protusions setting")
        layout.label(text="(Runs reqursively)")
        layout.label(text="If possible, avoid using on a high polycount base mesh")
        layout.label(text="(It can run out of memory and take a long time to compute)")

class VIEW3D_OT_tools_discombobulate(Operator):
    bl_idname = "discombobulate.ops"
    bl_label = "Discombobulator"
    bl_description = ("Easily add sci-fi details to a surface \n"
                      "Needs an existing active Mesh with Faces")
    bl_options = {"REGISTER", "UNDO"}

    executing = False

    # Protusions Buttons:
    repeatprot: IntProperty(
            name="Repeat protusions",
            description=("Make several layers of protusion \n"
                         "Use carefully, runs recursively the discombulator"),
            default=1, min=1, max=4  # set to 4 because it's 2**n reqursive
            )
    doprots: BoolProperty(
            name="Make protusions",
            description="Check if we want to add protusions to the mesh",
            default=True
            )
    subpolygon1: BoolProperty(
            name="1",
            default=True
            )
    subpolygon2: BoolProperty(
            name="2",
            default=True
            )
    subpolygon3: BoolProperty(
            name="3",
            default=True
            )
    subpolygon4: BoolProperty(
            name="4",
            default=True
            )
    polygonschangedpercent: FloatProperty(
            name="Polygon %",
            description="Percentage of changed polygons",
            default=1.0
            )
    minHeight: FloatProperty(
            name="Min height",
            description="Minimal height of the protusions",
            default=0.2
            )
    maxHeight: FloatProperty(
            name="Max height",
            description="Maximal height of the protusions",
            default=0.4
            )
    minTaper: FloatProperty(
            name="Min taper",
            description="Minimal height of the protusions",
            default=0.15, min=0.0, max=1.0,
            subtype='PERCENTAGE'
            )
    maxTaper: FloatProperty(
            name="Max taper",
            description="Maximal height of the protusions",
            default=0.35, min=0.0, max=1.0,
            subtype='PERCENTAGE'
            )
    # Doodads buttons:
    dodoodads: BoolProperty(
            name="Make doodads",
            description="Check if we want to generate doodads",
            default=False
            )
    mindoodads: IntProperty(
            name="Minimum doodads number",
            description="Ask for the minimum number of doodads to generate per polygon",
            default=1, min=0, max=50
            )
    maxdoodads: IntProperty(
            name="Maximum doodads number",
            description="Ask for the maximum number of doodads to generate per polygon",
            default=6, min=1, max=50
            )
    doodMinScale: FloatProperty(
            name="Scale min", description="Minimum scaling of doodad",
            default=0.5, min=0.0, max=1.0,
            subtype='PERCENTAGE'
            )
    doodMaxScale: FloatProperty(
            name="Scale max",
            description="Maximum scaling of doodad",
            default=1.0, min=0.0, max=1.0,
            subtype='PERCENTAGE'
            )
    # Materials buttons:
    sideProtMat: IntProperty(
            name="Side's prot mat",
            description="Material of protusion's sides",
            default=0, min=0
            )
    topProtMat: IntProperty(
            name="Prot's top mat",
            description="Material of protusion's top",
            default=0, min=0
            )

    @classmethod
    def poll(cls, context):
        return (context.active_object is not None and
                context.active_object.type == "MESH")

    def draw(self, context):
        layout = self.layout

        self.DISC_doodads = bpy.context.scene.discombobulator.DISC_doodads

        row = layout.row()
        row.menu("HELP_MT_discombobulator", icon="INFO")
        box = layout.box()
        box.label(text="Protusions settings")
        row = box.row()
        row.prop(self, "doprots")
        row = box.row()
        row.prop(self, "minHeight")
        row = box.row()
        row.prop(self, "maxHeight")
        row = box.row()
        row.prop(self, "minTaper")
        row = box.row()
        row.prop(self, "maxTaper")
        row = box.row()
        col1 = row.column(align=True)
        col1.prop(self, "subpolygon1")
        col2 = row.column(align=True)
        col2.prop(self, "subpolygon2")
        col3 = row.column(align=True)
        col3.prop(self, "subpolygon3")
        col4 = row.column(align=True)
        col4.prop(self, "subpolygon4")
        row = box.row()
        row.prop(self, "repeatprot")
        box = layout.box()
        box.label(text="Doodads settings")
        row = box.row()
        is_doodad = self.dodoodads
        row.prop(self, "dodoodads")

        row = box.row()
        row.enabled = is_doodad
        row.prop(self, "mindoodads")
        row = box.row()
        row.enabled = is_doodad
        row.prop(self, "maxdoodads")
        row = box.row()
        row.enabled = is_doodad
        oper = row.operator("object.discombobulate_set_doodad", text="Pick doodad")

        row = box.row()
        splits = row.split(factor = 0.5)
        splits.enabled = is_doodad
        splits.operator("object.discombobulate_unset_doodad",
                        text="Remove active doodad").remove_all = False
        splits.operator("object.discombobulate_unset_doodad",
                        text="Remove all doodads").remove_all = True

        col = box.column(align=True)
        doodle = len(self.DISC_doodads)

        col.enabled = (True if doodle > 0 else False)
        col.menu("OBJECT_MT_discombobulator_dodad_list",
                     text="List of saved Doodads ({})".format(doodle))

        box = layout.box()
        box.label(text="Materials settings")
        row = box.row()
        row.prop(self, "topProtMat")
        row = box.row()
        row.prop(self, "sideProtMat")

    def invoke(self, context, event):
        return context.window_manager.invoke_props_dialog(self, width=300)

    def check(self, context):
        return not self.executing

    def execute(self, context):
        self.executing = True
        i = 0
        while i < self.repeatprot:
            isLast = False
            if i == self.repeatprot - 1:
                isLast = True
            discombobulate(self, self.minHeight, self.maxHeight, self.minTaper, self.maxTaper, self.subpolygon1,
                           self.subpolygon2, self.subpolygon3, self.subpolygon4, self.mindoodads, self.maxdoodads,
                           self.repeatprot, self.sideProtMat, self.topProtMat, isLast)
            i += 1
        return {"FINISHED"}
        #bpy.ops.object.discombobulate("INVOKE_DEFAULT")



# Copyright 2018-2021 The glTF-Blender-IO authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bpy
import typing
from io_scene_gltf2.blender.exp.gltf2_blender_gather_cache import cached
from io_scene_gltf2.io.com import gltf2_io
from io_scene_gltf2.blender.exp import gltf2_blender_gather_texture
from io_scene_gltf2.blender.exp import gltf2_blender_search_node_tree
from io_scene_gltf2.blender.exp import gltf2_blender_get
from io_scene_gltf2.blender.exp.gltf2_blender_get import previous_node
from io_scene_gltf2.blender.exp.gltf2_blender_gather_sampler import detect_manual_uv_wrapping
from io_scene_gltf2.io.com.gltf2_io_extensions import Extension
from io_scene_gltf2.io.exp.gltf2_io_user_extensions import export_user_extensions


# blender_shader_sockets determine the texture and primary_socket determines
# the textranform and UVMap. Ex: when combining an ORM texture, for
# occlusion the primary_socket would be the occlusion socket, and
# blender_shader_sockets would be the (O,R,M) sockets.

def gather_texture_info(primary_socket, blender_shader_sockets, export_settings):
    return __gather_texture_info_helper(primary_socket, blender_shader_sockets, 'DEFAULT', export_settings)

def gather_material_normal_texture_info_class(primary_socket, blender_shader_sockets, export_settings):
    return __gather_texture_info_helper(primary_socket, blender_shader_sockets, 'NORMAL', export_settings)

def gather_material_occlusion_texture_info_class(primary_socket, blender_shader_sockets, export_settings):
    return __gather_texture_info_helper(primary_socket, blender_shader_sockets, 'OCCLUSION', export_settings)


@cached
def __gather_texture_info_helper(
        primary_socket: bpy.types.NodeSocket,
        blender_shader_sockets: typing.Tuple[bpy.types.NodeSocket],
        kind: str,
        export_settings):
    if not __filter_texture_info(primary_socket, blender_shader_sockets, export_settings):
        return None

    tex_transform, tex_coord = __gather_texture_transform_and_tex_coord(primary_socket, export_settings)

    fields = {
        'extensions': __gather_extensions(tex_transform, export_settings),
        'extras': __gather_extras(blender_shader_sockets, export_settings),
        'index': __gather_index(blender_shader_sockets, export_settings),
        'tex_coord': tex_coord,
    }

    if kind == 'DEFAULT':
        texture_info = gltf2_io.TextureInfo(**fields)

    elif kind == 'NORMAL':
        fields['scale'] = __gather_normal_scale(primary_socket, export_settings)
        texture_info = gltf2_io.MaterialNormalTextureInfoClass(**fields)

    elif kind == 'OCCLUSION':
        fields['strength'] = __gather_occlusion_strength(primary_socket, export_settings)
        texture_info = gltf2_io.MaterialOcclusionTextureInfoClass(**fields)

    if texture_info.index is None:
        return None

    export_user_extensions('gather_texture_info_hook', export_settings, texture_info, blender_shader_sockets)

    return texture_info


def __filter_texture_info(primary_socket, blender_shader_sockets, export_settings):
    if primary_socket is None:
        return False
    if __get_tex_from_socket(primary_socket) is None:
        return False
    if not blender_shader_sockets:
        return False
    if not all([elem is not None for elem in blender_shader_sockets]):
        return False
    if any([__get_tex_from_socket(socket) is None for socket in blender_shader_sockets]):
        # sockets do not lead to a texture --> discard
        return False

    return True


def __gather_extensions(texture_transform, export_settings):
    if texture_transform is None:
        return None
    extension = Extension("KHR_texture_transform", texture_transform)
    return {"KHR_texture_transform": extension}


def __gather_extras(blender_shader_sockets, export_settings):
    return None


# MaterialNormalTextureInfo only
def __gather_normal_scale(primary_socket, export_settings):
    result = gltf2_blender_search_node_tree.from_socket(
        primary_socket,
        gltf2_blender_search_node_tree.FilterByType(bpy.types.ShaderNodeNormalMap))
    if not result:
        return None
    strengthInput = result[0].shader_node.inputs['Strength']
    if not strengthInput.is_linked and strengthInput.default_value != 1:
        return strengthInput.default_value
    return None


# MaterialOcclusionTextureInfo only
def __gather_occlusion_strength(primary_socket, export_settings):
    # Look for a MixRGB node that mixes with pure white in front of
    # primary_socket. The mix factor gives the occlusion strength.
    node = gltf2_blender_get.previous_node(primary_socket)
    if node and node.type == 'MIX_RGB' and node.blend_type == 'MIX':
        fac = gltf2_blender_get.get_const_from_socket(node.inputs['Fac'], kind='VALUE')
        col1 = gltf2_blender_get.get_const_from_socket(node.inputs['Color1'], kind='RGB')
        col2 = gltf2_blender_get.get_const_from_socket(node.inputs['Color2'], kind='RGB')
        if fac is not None:
            if col1 == [1, 1, 1] and col2 is None:
                return fac
            if col1 is None and col2 == [1, 1, 1]:
                return 1.0 - fac  # reversed for reversed inputs

    return None


def __gather_index(blender_shader_sockets, export_settings):
    # We just put the actual shader into the 'index' member
    return gltf2_blender_gather_texture.gather_texture(blender_shader_sockets, export_settings)


def __gather_texture_transform_and_tex_coord(primary_socket, export_settings):
    # We're expecting
    #
    #     [UV Map] => [Mapping] => [UV Wrapping] => [Texture Node] => ... => primary_socket
    #
    # The [UV Wrapping] is for wrap modes like MIRROR that use nodes,
    # [Mapping] is for KHR_texture_transform, and [UV Map] is for texCoord.
    blender_shader_node = __get_tex_from_socket(primary_socket).shader_node

    # Skip over UV wrapping stuff (it goes in the sampler)
    result = detect_manual_uv_wrapping(blender_shader_node)
    if result:
        node = previous_node(result['next_socket'])
    else:
        node = previous_node(blender_shader_node.inputs['Vector'])

    texture_transform = None
    if node and node.type == 'MAPPING':
        texture_transform = gltf2_blender_get.get_texture_transform_from_mapping_node(node)
        node = previous_node(node.inputs['Vector'])

    texcoord_idx = 0
    if node and node.type == 'UVMAP' and node.uv_map:
        # Try to gather map index.
        for blender_mesh in bpy.data.meshes:
            i = blender_mesh.uv_layers.find(node.uv_map)
            if i >= 0:
                texcoord_idx = i
                break

    return texture_transform, texcoord_idx or None


def __get_tex_from_socket(socket):
    result = gltf2_blender_search_node_tree.from_socket(
        socket,
        gltf2_blender_search_node_tree.FilterByType(bpy.types.ShaderNodeTexImage))
    if not result:
        return None
    if result[0].shader_node.image is None:
        return None
    return result[0]


def check_same_size_images(
    blender_shader_sockets: typing.Tuple[bpy.types.NodeSocket],
) -> bool:
    """Check that all sockets leads to images of the same size."""
    if not blender_shader_sockets or not all(blender_shader_sockets):
        return False

    sizes = set()
    for socket in blender_shader_sockets:
        tex = __get_tex_from_socket(socket)
        if tex is None:
            return False
        size = tex.shader_node.image.size
        sizes.add((size[0], size[1]))

    return len(sizes) == 1



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>

'''
Define the command 'print_struct_c99' for gdb,
useful for creating a literal for a nested runtime struct.

Example use:

   (gdb) source source/tools/utils/gdb_struct_repr_c99.py
   (gdb) print_struct_c99 scene->toolsettings
'''

class PrintStructC99(gdb.Command):
    def __init__(self):
        super(PrintStructC99, self).__init__(
            "print_struct_c99",
            gdb.COMMAND_USER,
        )

    def get_count_heading(self, string):
        for i, s in enumerate(string):
            if s != ' ':
                break
        return i

    def extract_typename(self, string):
        first_line = string.split('\n')[0]
        return first_line.split('=')[1][:-1].strip()

    def invoke(self, arg, from_tty):
        ret_ptype = gdb.execute('ptype {}'.format(arg), to_string=True)
        tname = self.extract_typename(ret_ptype)
        print('{} {} = {{'.format(tname, arg))
        r = gdb.execute('p {}'.format(arg), to_string=True)
        r = r.split('\n')
        for rr in r[1:]:
            if '=' not in rr:
                print(rr)
                continue
            hs = self.get_count_heading(rr)
            rr_s = rr.strip().split('=', 1)
            rr_rval = rr_s[1].strip()
            print(' ' * hs + '.' + rr_s[0] + '= ' + rr_rval)


print('Running GDB from: %s\n' % (gdb.PYTHONDIR))
gdb.execute("set print pretty")
gdb.execute('set pagination off')
gdb.execute('set print repeats 0')
gdb.execute('set print elements unlimited')
# instantiate
PrintStructC99()



### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##### END GPL LICENSE BLOCK #####
import bpy

from mathutils import (
    Vector,
    Matrix,
    )
from mathutils.geometry import intersect_point_line
from .drawing_utilities import SnapDrawn
from .common_utilities import (
    convert_distance,
    get_units_info,
    location_3d_to_region_2d,
    )


class SnapNavigation():
    __slots__ = (
        'use_ndof',
        '_rotate',
        '_move',
        '_zoom',
        '_ndof_all',
        '_ndof_orbit',
        '_ndof_orbit_zoom',
        '_ndof_pan')


    @staticmethod
    def debug_key(key):
        for member in dir(key):
            print(member, getattr(key, member))

    @staticmethod
    def convert_to_flag(shift, ctrl, alt):
        return (shift << 0) | (ctrl << 1) | (alt << 2)

    def __init__(self, context, use_ndof):
        # TO DO:
        # 'View Orbit', 'View Pan', 'NDOF Orbit View', 'NDOF Pan View'
        self.use_ndof = use_ndof and context.preferences.inputs.use_ndof

        self._rotate = set()
        self._move = set()
        self._zoom = set()

        if self.use_ndof:
            self._ndof_all = set()
            self._ndof_orbit = set()
            self._ndof_orbit_zoom = set()
            self._ndof_pan = set()

        for key in context.window_manager.keyconfigs.user.keymaps['3D View'].keymap_items:
            if key.idname == 'view3d.rotate':
                self._rotate.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type, key.value))
            elif key.idname == 'view3d.move':
                self._move.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type, key.value))
            elif key.idname == 'view3d.zoom':
                if key.type == 'WHEELINMOUSE':
                    self._zoom.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), 'WHEELUPMOUSE', key.value, key.properties.delta))
                elif key.type == 'WHEELOUTMOUSE':
                    self._zoom.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), 'WHEELDOWNMOUSE', key.value, key.properties.delta))
                else:
                    self._zoom.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type, key.value, key.properties.delta))

            elif self.use_ndof:
                if key.idname == 'view3d.ndof_all':
                    self._ndof_all.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type))
                elif key.idname == 'view3d.ndof_orbit':
                    self._ndof_orbit.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type))
                elif key.idname == 'view3d.ndof_orbit_zoom':
                    self._ndof_orbit_zoom.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type))
                elif key.idname == 'view3d.ndof_pan':
                    self._ndof_pan.add((self.convert_to_flag(key.shift, key.ctrl, key.alt), key.type))


    def run(self, context, event, snap_location):
        evkey = (self.convert_to_flag(event.shift, event.ctrl, event.alt), event.type, event.value)

        if evkey in self._rotate:
            if snap_location:
                bpy.ops.view3d.rotate_custom_pivot('INVOKE_DEFAULT', pivot=snap_location)
            else:
                bpy.ops.view3d.rotate('INVOKE_DEFAULT', use_cursor_init=True)
            return True

        if evkey in self._move:
            bpy.ops.view3d.move('INVOKE_DEFAULT')
            return True

        for key in self._zoom:
            if evkey == key[0:3]:
                if key[3]:
                    if snap_location:
                        bpy.ops.view3d.zoom_custom_target('INVOKE_DEFAULT', delta=key[3], target=snap_location)
                    else:
                        bpy.ops.view3d.zoom('INVOKE_DEFAULT', delta=key[3])
                else:
                    bpy.ops.view3d.zoom('INVOKE_DEFAULT')
                return True

        if self.use_ndof:
            ndofkey = evkey[:2]
            if ndofkey in self._ndof_all:
                bpy.ops.view3d.ndof_all('INVOKE_DEFAULT')
                return True
            if ndofkey in self._ndof_orbit:
                bpy.ops.view3d.ndof_orbit('INVOKE_DEFAULT')
                return True
            if ndofkey in self._ndof_orbit_zoom:
                bpy.ops.view3d.ndof_orbit_zoom('INVOKE_DEFAULT')
                return True
            if ndofkey in self._ndof_pan:
                bpy.ops.view3d.ndof_pan('INVOKE_DEFAULT')
                return True

        return False


class CharMap:
    __slots__ = (
        'unit_system',
        'uinfo',
        'length_entered',
        'length_entered_value',
        'line_pos')

    ascii = {
        ".", ",", "-", "+", "1", "2", "3",
        "4", "5", "6", "7", "8", "9", "0",
        "c", "m", "d", "k", "h", "a",
        " ", "/", "*", "'", "\""
        # "="
        }
    type = {
        'BACK_SPACE', 'DEL',
        'LEFT_ARROW', 'RIGHT_ARROW'
        }

    def __init__(self, context):
        scale = context.scene.unit_settings.scale_length
        separate_units = context.scene.unit_settings.use_separate
        self.unit_system = context.scene.unit_settings.system
        self.uinfo = get_units_info(scale, self.unit_system, separate_units)

        self.clear()

    def modal_(self, context, event):
        if event.value == 'PRESS':
            type = event.type
            ascii = event.ascii
            if (type in self.type) or (ascii in self.ascii):
                if ascii:
                    pos = self.line_pos
                    if ascii == ",":
                        ascii = "."
                    self.length_entered = self.length_entered[:pos] + ascii + self.length_entered[pos:]
                    self.line_pos += 1

                if self.length_entered:
                    pos = self.line_pos
                    if type == 'BACK_SPACE':
                        self.length_entered = self.length_entered[:pos - 1] + self.length_entered[pos:]
                        self.line_pos -= 1

                    elif type == 'DEL':
                        self.length_entered = self.length_entered[:pos] + self.length_entered[pos + 1:]

                    elif type == 'LEFT_ARROW':
                        self.line_pos = (pos - 1) % (len(self.length_entered) + 1)

                    elif type == 'RIGHT_ARROW':
                        self.line_pos = (pos + 1) % (len(self.length_entered) + 1)

                    try:
                        self.length_entered_value = bpy.utils.units.to_value(self.unit_system, 'LENGTH', self.length_entered)
                    except:  # ValueError:
                        self.length_entered_value = 0.0 #invalid
                        #self.report({'INFO'}, "Operation not supported yet")
                else:
                    self.length_entered_value = 0.0

                return True

        return False

    def get_converted_length_str(self, length):
        if self.length_entered:
            pos = self.line_pos
            ret = self.length_entered[:pos] + '|' + self.length_entered[pos:]
        else:
            ret = convert_distance(length, self.uinfo)

        return ret

    def clear(self):
        self.length_entered = ''
        self.length_entered_value = 0.0
        self.line_pos = 0


class Constrain:
    def __init__(self, peferences, scene, obj):
        self.last_type = None
        self.last_vec = None
        self.rotMat = None
        self.preferences = peferences
        trans_orient = scene.transform_orientation_slots[0]
        self.orientation = [None, None]
        if trans_orient.type == 'LOCAL':
            self.orientation[0] = obj.matrix_world.to_3x3().transposed()
            self.orientation[1] = Matrix.Identity(3)
        else:
            self.orientation[0] = Matrix.Identity(3)
            self.orientation[1] = obj.matrix_world.to_3x3().transposed()

        self.orientation_id = 0
        self.center = Vector((0.0, 0.0, 0.0))
        self.center_2d = Vector((0.0, 0.0))
        self.projected_vecs = Matrix(([0.0, 0.0], [0.0, 0.0], [0.0, 0.0]))

    def _constrain_set(self, mcursor):
        vec = (mcursor - self.center_2d)
        vec.normalize()

        dot_x = abs(vec.dot(self.projected_vecs[0]))
        dot_y = abs(vec.dot(self.projected_vecs[1]))
        dot_z = abs(vec.dot(self.projected_vecs[2]))

        if dot_x > dot_y and dot_x > dot_z:
            vec = self.orientation[self.orientation_id][0]
            type = 'X'

        elif dot_y > dot_x and dot_y > dot_z:
            vec = self.orientation[self.orientation_id][1]
            type = 'Y'

        else: # dot_z > dot_y and dot_z > dot_x:
            vec = self.orientation[self.orientation_id][2]
            type = 'Z'

        return vec, type

    def modal(self, event, shift_callback):
        type = event.type
        if self.last_type == type:
            self.orientation_id += 1

        if type == 'X':
            if self.orientation_id < 2:
                self.last_vec = self.orientation[self.orientation_id][0]
            else:
                self.orientation_id = 0
                self.last_vec = type = None
        elif type == 'Y':
            if self.orientation_id < 2:
                self.last_vec = self.orientation[self.orientation_id][1]
            else:
                self.orientation_id = 0
                self.last_vec = type = None
        elif type == 'Z':
            if self.orientation_id < 2:
                self.last_vec = self.orientation[self.orientation_id][2]
            else:
                self.orientation_id = 0
                self.last_vec = type = None
        elif shift_callback and type in {'RIGHT_SHIFT', 'LEFT_SHIFT'}:
            if self.orientation_id < 1:
                type = 'shift'
                self.last_vec = shift_callback()
            else:
                self.orientation_id = 0
                self.last_vec = type = None
        else:
            return False

        self.preferences.auto_constrain = False
        self.last_type = type
        return True

    def toogle(self):
        self.rotMat = None # update
        if self.preferences.auto_constrain:
            self.orientation_id = (self.orientation_id + 1) % 2
            self.preferences.auto_constrain = self.orientation_id != 0
        else:
            self.preferences.auto_constrain = True

    def update(self, region, rv3d, mcursor, center):
        if rv3d.view_matrix != self.rotMat or self.center != center:
            self.rotMat = rv3d.view_matrix.copy()

            self.center = center.copy()
            self.center_2d = location_3d_to_region_2d(region, rv3d, self.center)

            vec = self.center + self.orientation[self.orientation_id][0]
            self.projected_vecs[0] = location_3d_to_region_2d(region, rv3d, vec) - self.center_2d
            vec = self.center + self.orientation[self.orientation_id][1]
            self.projected_vecs[1] = location_3d_to_region_2d(region, rv3d, vec) - self.center_2d
            vec = self.center + self.orientation[self.orientation_id][2]
            self.projected_vecs[2] = location_3d_to_region_2d(region, rv3d, vec) - self.center_2d

            self.projected_vecs[0].normalize()
            self.projected_vecs[1].normalize()
            self.projected_vecs[2].normalize()

        return self._constrain_set(mcursor)


class SnapUtilities:
    """
    __slots__ = (
        "sctx",
        "draw_cache",
        "outer_verts",
        "unit_system",
        "rd",
        "obj",
        "bm",
        "geom",
        "type",
        "location",
        "preferences",
        "normal",
        "snap_vert",
        "snap_edge",
        "snap_face",
        "incremental",
    )
    """

    constrain_keys = {
        'X': Vector((1,0,0)),
        'Y': Vector((0,1,0)),
        'Z': Vector((0,0,1)),
        'RIGHT_SHIFT': 'shift',
        'LEFT_SHIFT': 'shift',
        }

    snapwidgets = []
    constrain = None

    @staticmethod
    def set_contrain(context, key):
        widget = SnapUtilities.snapwidgets[-1] if SnapUtilities.snapwidgets else None
        if SnapUtilities.constrain == key:
            SnapUtilities.constrain = None
            if hasattr(widget, "get_normal"):
                widget.get_normal(context)
            return

        if hasattr(widget, "normal"):
            if key == 'shift':
                import bmesh
                if isinstance(widget.geom, bmesh.types.BMEdge):
                    verts = widget.geom.verts
                    widget.normal = verts[1].co - verts[0].co
                    widget.normal.normalise()
                else:
                    return
            else:
                widget.normal = SnapUtilities.constrain_keys[key]

        SnapUtilities.constrain = key


    def snap_context_update_and_return_moving_objects(self, context):
        moving_objects = set()
        moving_snp_objects = set()
        children = set()
        for obj in context.view_layer.objects.selected:
            moving_objects.add(obj)

        temp_children = set()
        for obj in context.visible_objects:
            temp_children.clear()
            while obj.parent is not None:
                temp_children.add(obj)
                parent = obj.parent
                if parent in moving_objects:
                    children.update(temp_children)
                    temp_children.clear()
                obj = parent

        del temp_children

        moving_objects.difference_update(children)

        self.sctx.clear_snap_objects(True)

        for obj in context.visible_objects:
            is_moving = obj in moving_objects or obj in children
            snap_obj = self.sctx.add_obj(obj, obj.matrix_world)
            if is_moving:
                moving_snp_objects.add(snap_obj)

            if obj.instance_type == 'COLLECTION':
                mat = obj.matrix_world.copy()
                for ob in obj.instance_collection.objects:
                    snap_obj = self.sctx.add_obj(ob, mat @ ob.matrix_world)
                    if is_moving:
                        moving_snp_objects.add(snap_obj)

        del children
        return moving_objects, moving_snp_objects


    def snap_context_update(self, context):
        def visible_objects_and_duplis():
            if self.preferences.outer_verts:
                for obj in context.visible_objects:
                    yield (obj, obj.matrix_world)

                    if obj.instance_type == 'COLLECTION':
                        mat = obj.matrix_world.copy()
                        for ob in obj.instance_collection.objects:
                            yield (ob, mat @ ob.matrix_world)
            else:
                for obj in context.objects_in_mode_unique_data:
                    yield (obj, obj.matrix_world)

        self.sctx.clear_snap_objects(True)

        for obj, matrix in visible_objects_and_duplis():
            self.sctx.add_obj(obj, matrix)


    def snap_context_init(self, context, snap_edge_and_vert=True):
        from .snap_context_l import global_snap_context_get

        #Create Snap Context
        self.sctx = global_snap_context_get(context.evaluated_depsgraph_get(), context.region, context.space_data)
        self.sctx.set_pixel_dist(12)
        self.sctx.use_clip_planes(True)

        if SnapUtilities.snapwidgets:
            widget = SnapUtilities.snapwidgets[-1]

            self.obj = widget.snap_obj.data[0] if widget.snap_obj else context.active_object
            self.bm = widget.bm
            self.geom = widget.geom
            self.type = widget.type
            self.location = widget.location
            self.preferences = widget.preferences
            self.draw_cache = widget.draw_cache
            if hasattr(widget, "normal"):
                self.normal = widget.normal

        else:
            #init these variables to avoid errors
            self.obj = context.active_object
            self.bm = None
            self.geom = None
            self.type = 'OUT'
            self.location = Vector()

            preferences = context.preferences.addons[__package__].preferences
            self.preferences = preferences
            #Init DrawCache
            self.draw_cache = SnapDrawn(
                preferences.out_color,
                preferences.face_color,
                preferences.edge_color,
                preferences.vert_color,
                preferences.center_color,
                preferences.perpendicular_color,
                preferences.constrain_shift_color,
                tuple(context.preferences.themes[0].user_interface.axis_x) + (1.0,),
                tuple(context.preferences.themes[0].user_interface.axis_y) + (1.0,),
                tuple(context.preferences.themes[0].user_interface.axis_z) + (1.0,))

        self.snap_vert = self.snap_edge = snap_edge_and_vert

        shading = context.space_data.shading
        self.snap_face = not (snap_edge_and_vert and (shading.show_xray or shading.type == 'WIREFRAME'))

        self.sctx.set_snap_mode(self.snap_vert, self.snap_edge, self.snap_face)

        #Configure the unit of measure
        unit_system = context.scene.unit_settings.system
        scale = context.scene.unit_settings.scale_length
        scale /= context.space_data.overlay.grid_scale
        self.rd = bpy.utils.units.to_value(unit_system, 'LENGTH', str(1 / scale))

        self.incremental = bpy.utils.units.to_value(unit_system, 'LENGTH', str(self.preferences.incremental))

    def snap_to_grid(self):
        if self.type == 'OUT' and self.preferences.increments_grid:
            loc = self.location / self.rd
            self.location = Vector((round(loc.x),
                                    round(loc.y),
                                    round(loc.z))) * self.rd

    def snap_context_free(self):
        self.sctx = None
        del self.sctx

        del self.bm
        del self.draw_cache
        del self.geom
        del self.location
        del self.rd
        del self.snap_face
        del self.snap_obj
        del self.type

        del self.preferences

        SnapUtilities.constrain = None



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>

import bpy
from bpy.types import Operator
from bpy.props import StringProperty


class EditExternally(Operator):
    """Edit image in an external application"""
    bl_idname = "image.external_edit"
    bl_label = "Image Edit Externally"
    bl_options = {'REGISTER'}

    filepath: StringProperty(
        subtype='FILE_PATH',
    )

    @staticmethod
    def _editor_guess(context):
        import sys

        image_editor = context.preferences.filepaths.image_editor

        # use image editor in the preferences when available.
        if not image_editor:
            if sys.platform[:3] == "win":
                image_editor = ["start"]  # not tested!
            elif sys.platform == "darwin":
                image_editor = ["open"]
            else:
                image_editor = ["gimp"]
        else:
            if sys.platform == "darwin":
                # blender file selector treats .app as a folder
                # and will include a trailing backslash, so we strip it.
                image_editor.rstrip('\\')
                image_editor = ["open", "-a", image_editor]
            else:
                image_editor = [image_editor]

        return image_editor

    def execute(self, context):
        import os
        import subprocess

        filepath = self.filepath

        if not filepath:
            self.report({'ERROR'}, "Image path not set")
            return {'CANCELLED'}

        if not os.path.exists(filepath) or not os.path.isfile(filepath):
            self.report({'ERROR'},
                        "Image path %r not found, image may be packed or "
                        "unsaved" % filepath)
            return {'CANCELLED'}

        cmd = self._editor_guess(context) + [filepath]

        try:
            subprocess.Popen(cmd)
        except:
            import traceback
            traceback.print_exc()
            self.report({'ERROR'},
                        "Image editor could not be launched, ensure that "
                        "the path in User Preferences > File is valid, and Blender has rights to launch it")

            return {'CANCELLED'}

        return {'FINISHED'}

    def invoke(self, context, _event):
        import os
        sd = context.space_data
        try:
            image = sd.image
        except AttributeError:
            self.report({'ERROR'}, "Context incorrect, image not found")
            return {'CANCELLED'}

        if image.packed_file:
            self.report({'ERROR'}, "Image is packed, unpack before editing")
            return {'CANCELLED'}

        if sd.type == 'IMAGE_EDITOR':
            filepath = image.filepath_from_user(image_user=sd.image_user)
        else:
            filepath = image.filepath

        filepath = bpy.path.abspath(filepath, library=image.library)

        self.filepath = os.path.normpath(filepath)
        self.execute(context)

        return {'FINISHED'}


class ProjectEdit(Operator):
    """Edit a snapshot of the 3D Viewport in an external image editor"""
    bl_idname = "image.project_edit"
    bl_label = "Project Edit"
    bl_options = {'REGISTER'}

    _proj_hack = [""]

    def execute(self, context):
        import os

        EXT = "png"  # could be made an option but for now ok

        for image in bpy.data.images:
            image.tag = True

        # opengl buffer may fail, we can't help this, but best report it.
        try:
            bpy.ops.paint.image_from_view()
        except RuntimeError as err:
            self.report({'ERROR'}, str(err))
            return {'CANCELLED'}

        image_new = None
        for image in bpy.data.images:
            if not image.tag:
                image_new = image
                break

        if not image_new:
            self.report({'ERROR'}, "Could not make new image")
            return {'CANCELLED'}

        filepath = os.path.basename(bpy.data.filepath)
        filepath = os.path.splitext(filepath)[0]
        # fixes <memory> rubbish, needs checking
        # filepath = bpy.path.clean_name(filepath)

        if bpy.data.is_saved:
            filepath = "//" + filepath
        else:
            filepath = os.path.join(bpy.app.tempdir, "project_edit")

        obj = context.object

        if obj:
            filepath += "_" + bpy.path.clean_name(obj.name)

        filepath_final = filepath + "." + EXT
        i = 0

        while os.path.exists(bpy.path.abspath(filepath_final)):
            filepath_final = filepath + ("%.3d.%s" % (i, EXT))
            i += 1

        image_new.name = bpy.path.basename(filepath_final)
        ProjectEdit._proj_hack[0] = image_new.name

        image_new.filepath_raw = filepath_final  # TODO, filepath raw is crummy
        image_new.file_format = 'PNG'
        image_new.save()

        filepath_final = bpy.path.abspath(filepath_final)

        try:
            bpy.ops.image.external_edit(filepath=filepath_final)
        except RuntimeError as re:
            self.report({'ERROR'}, str(re))

        return {'FINISHED'}


class ProjectApply(Operator):
    """Project edited image back onto the object"""
    bl_idname = "image.project_apply"
    bl_label = "Project Apply"
    bl_options = {'REGISTER'}

    def execute(self, _context):
        image_name = ProjectEdit._proj_hack[0]  # TODO, deal with this nicer

        try:
            image = bpy.data.images[image_name, None]
        except KeyError:
            import traceback
            traceback.print_exc()
            self.report({'ERROR'}, "Could not find image '%s'" % image_name)
            return {'CANCELLED'}

        image.reload()
        bpy.ops.paint.project_image(image=image_name)

        return {'FINISHED'}


classes = (
    EditExternally,
    ProjectApply,
    ProjectEdit,
)



import mathutils
from math import radians

vec = mathutils.Vector((1.0, 2.0, 3.0))

mat_rot = mathutils.Matrix.Rotation(radians(90.0), 4, 'X')
mat_trans = mathutils.Matrix.Translation(vec)

mat = mat_trans @ mat_rot
mat.invert()

mat3 = mat.to_3x3()
quat1 = mat.to_quaternion()
quat2 = mat3.to_quaternion()

quat_diff = quat1.rotation_difference(quat2)

print(quat_diff.angle)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>


def url_prefill_from_blender(addon_info=None):
    import bpy
    import bgl
    import struct
    import platform
    import urllib.parse
    import io

    fh = io.StringIO()

    fh.write("**System Information**\n")
    fh.write(
        "Operating system: %s %d Bits\n" % (
            platform.platform(),
            struct.calcsize("P") * 8,
        )
    )
    fh.write(
        "Graphics card: %s %s %s\n" % (
            bgl.glGetString(bgl.GL_RENDERER),
            bgl.glGetString(bgl.GL_VENDOR),
            bgl.glGetString(bgl.GL_VERSION),
        )
    )
    fh.write(
        "\n"
        "**Blender Version**\n"
    )
    fh.write(
        "Broken: version: %s, branch: %s, commit date: %s %s, hash: `rB%s`\n" % (
            bpy.app.version_string,
            bpy.app.build_branch.decode('utf-8', 'replace'),
            bpy.app.build_commit_date.decode('utf-8', 'replace'),
            bpy.app.build_commit_time.decode('utf-8', 'replace'),
            bpy.app.build_hash.decode('ascii'),
        )
    )
    fh.write(
        "Worked: (newest version of Blender that worked as expected)\n"
    )
    if addon_info:
        fh.write(
            "\n"
            "**Addon Information**\n"
        )
        fh.write(addon_info)

    fh.write(
        "\n"
        "**Short description of error**\n"
        "[Please fill out a short description of the error here]\n"
        "\n"
        "**Exact steps for others to reproduce the error**\n"
        "[Please describe the exact steps needed to reproduce the issue]\n"
        "[Based on the default startup or an attached .blend file (as simple as possible)]\n"
        "\n"
    )

    fh.seek(0)

    form_number = 2 if addon_info else 1
    return (
        "https://developer.blender.org/maniphest/task/edit/form/%i?description=" % form_number +
        urllib.parse.quote(fh.read())
    )



# <pep8-80 compliant>

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

__author__ = "Nutti <nutti.metro@gmail.com>"
__status__ = "production"
__version__ = "6.5"
__date__ = "6 Mar 2021"

import bpy.utils

from ..op.copy_paste_uv import (
    MUV_MT_CopyPasteUV_CopyUV,
    MUV_MT_CopyPasteUV_PasteUV,
    MUV_MT_CopyPasteUV_SelSeqCopyUV,
    MUV_MT_CopyPasteUV_SelSeqPasteUV,
)
from ..op.transfer_uv import (
    MUV_OT_TransferUV_CopyUV,
    MUV_OT_TransferUV_PasteUV,
)
from ..op.uvw import (
    MUV_OT_UVW_BoxMap,
    MUV_OT_UVW_BestPlanerMap,
)
from ..op.preserve_uv_aspect import MUV_OT_PreserveUVAspect
from ..op.texture_lock import (
    MUV_OT_TextureLock_Lock,
    MUV_OT_TextureLock_Unlock,
)
from ..op.texture_wrap import (
    MUV_OT_TextureWrap_Refer,
    MUV_OT_TextureWrap_Set,
)
from ..op.world_scale_uv import (
    MUV_OT_WorldScaleUV_Measure,
    MUV_OT_WorldScaleUV_ApplyManual,
    MUV_OT_WorldScaleUV_ApplyScalingDensity,
    MUV_OT_WorldScaleUV_ApplyProportionalToMesh,
)
from ..op.texture_projection import MUV_OT_TextureProjection_Project
from ..utils.bl_class_registry import BlClassRegistry


@BlClassRegistry()
class MUV_MT_CopyPasteUV(bpy.types.Menu):
    """
    Menu class: Master menu of Copy/Paste UV coordinate
    """

    bl_idname = "MUV_MT_CopyPasteUV"
    bl_label = "Copy/Paste UV"
    bl_description = "Copy and Paste UV coordinate"

    def draw(self, _):
        layout = self.layout

        layout.label(text="Default")
        layout.menu(MUV_MT_CopyPasteUV_CopyUV.bl_idname, text="Copy")
        layout.menu(MUV_MT_CopyPasteUV_PasteUV.bl_idname, text="Paste")

        layout.separator()

        layout.label(text="Selection Sequence")
        layout.menu(MUV_MT_CopyPasteUV_SelSeqCopyUV.bl_idname, text="Copy")
        layout.menu(MUV_MT_CopyPasteUV_SelSeqPasteUV.bl_idname, text="Paste")


@BlClassRegistry()
class MUV_MT_TransferUV(bpy.types.Menu):
    """
    Menu class: Master menu of Transfer UV coordinate
    """

    bl_idname = "MUV_MT_TransferUV"
    bl_label = "Transfer UV"
    bl_description = "Transfer UV coordinate"

    def draw(self, context):
        layout = self.layout
        sc = context.scene

        layout.operator(MUV_OT_TransferUV_CopyUV.bl_idname, text="Copy")
        ops = layout.operator(MUV_OT_TransferUV_PasteUV.bl_idname,
                              text="Paste")
        ops.invert_normals = sc.muv_transfer_uv_invert_normals
        ops.copy_seams = sc.muv_transfer_uv_copy_seams


@BlClassRegistry()
class MUV_MT_TextureLock(bpy.types.Menu):
    """
    Menu class: Master menu of Texture Lock
    """

    bl_idname = "MUV_MT_TextureLock"
    bl_label = "Texture Lock"
    bl_description = "Lock texture when vertices of mesh (Preserve UV)"

    def draw(self, context):
        layout = self.layout
        sc = context.scene

        layout.label(text="Normal Mode")
        layout.operator(
            MUV_OT_TextureLock_Lock.bl_idname,
            text="Lock"
            if not MUV_OT_TextureLock_Lock.is_ready(context)
            else "ReLock")
        ops = layout.operator(MUV_OT_TextureLock_Unlock.bl_idname,
                              text="Unlock")
        ops.connect = sc.muv_texture_lock_connect

        layout.separator()

        layout.label(text="Interactive Mode")
        layout.prop(sc, "muv_texture_lock_lock", text="Lock")


@BlClassRegistry()
class MUV_MT_WorldScaleUV(bpy.types.Menu):
    """
    Menu class: Master menu of world scale UV
    """

    bl_idname = "MUV_MT_WorldScaleUV"
    bl_label = "World Scale UV"
    bl_description = ""

    def draw(self, context):
        layout = self.layout
        sc = context.scene

        layout.operator(MUV_OT_WorldScaleUV_Measure.bl_idname, text="Measure")

        ops = layout.operator(MUV_OT_WorldScaleUV_ApplyManual.bl_idname,
                              text="Apply (Manual)")
        ops.show_dialog = True

        ops = layout.operator(
            MUV_OT_WorldScaleUV_ApplyScalingDensity.bl_idname,
            text="Apply (Same Desity)")
        ops.src_density = sc.muv_world_scale_uv_src_density
        ops.same_density = True
        ops.show_dialog = True

        ops = layout.operator(
            MUV_OT_WorldScaleUV_ApplyScalingDensity.bl_idname,
            text="Apply (Scaling Desity)")
        ops.src_density = sc.muv_world_scale_uv_src_density
        ops.same_density = False
        ops.show_dialog = True

        ops = layout.operator(
            MUV_OT_WorldScaleUV_ApplyProportionalToMesh.bl_idname,
            text="Apply (Proportional to Mesh)")
        ops.src_density = sc.muv_world_scale_uv_src_density
        ops.src_uv_area = sc.muv_world_scale_uv_src_uv_area
        ops.src_mesh_area = sc.muv_world_scale_uv_src_mesh_area
        ops.show_dialog = True


@BlClassRegistry()
class MUV_MT_TextureWrap(bpy.types.Menu):
    """
    Menu class: Master menu of Texture Wrap
    """

    bl_idname = "MUV_MT_TextureWrap"
    bl_label = "Texture Wrap"
    bl_description = ""

    def draw(self, _):
        layout = self.layout

        layout.operator(MUV_OT_TextureWrap_Refer.bl_idname, text="Refer")
        layout.operator(MUV_OT_TextureWrap_Set.bl_idname, text="Set")


@BlClassRegistry()
class MUV_MT_UVW(bpy.types.Menu):
    """
    Menu class: Master menu of UVW
    """

    bl_idname = "MUV_MT_UVW"
    bl_label = "UVW"
    bl_description = ""

    def draw(self, context):
        layout = self.layout
        sc = context.scene

        ops = layout.operator(MUV_OT_UVW_BoxMap.bl_idname, text="Box")
        ops.assign_uvmap = sc.muv_uvw_assign_uvmap

        ops = layout.operator(MUV_OT_UVW_BestPlanerMap.bl_idname,
                              text="Best Planner")
        ops.assign_uvmap = sc.muv_uvw_assign_uvmap


@BlClassRegistry()
class MUV_MT_PreserveUVAspect(bpy.types.Menu):
    """
    Menu class: Master menu of Preserve UV Aspect
    """

    bl_idname = "MUV_MT_PreserveUVAspect"
    bl_label = "Preserve UV Aspect"
    bl_description = ""

    def draw(self, context):
        layout = self.layout
        sc = context.scene

        for key in bpy.data.images.keys():
            ops = layout.operator(MUV_OT_PreserveUVAspect.bl_idname, text=key)
            ops.dest_img_name = key
            ops.origin = sc.muv_preserve_uv_aspect_origin


@BlClassRegistry()
class MUV_MT_TextureProjection(bpy.types.Menu):
    """
    Menu class: Master menu of Texture Projection
    """

    bl_idname = "MUV_MT_TextureProjection"
    bl_label = "Texture Projection"
    bl_description = ""

    def draw(self, context):
        layout = self.layout
        sc = context.scene

        layout.prop(sc, "muv_texture_projection_enable",
                    text="Texture Projection")
        layout.operator(MUV_OT_TextureProjection_Project.bl_idname,
                        text="Project")



# Copyright 2018-2021 The glTF-Blender-IO authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bpy

def simulate_stash(obj, track_name, action, start_frame=None):
    # Simulate stash :
    # * add a track
    # * add an action on track
    # * lock & mute the track
    if not obj.animation_data:
        obj.animation_data_create()
    tracks = obj.animation_data.nla_tracks
    new_track = tracks.new(prev=None)
    new_track.name = track_name
    if start_frame is None:
        start_frame = bpy.context.scene.frame_start
    _strip = new_track.strips.new(action.name, start_frame, action)
    new_track.lock = True
    new_track.mute = True

def restore_animation_on_object(obj, anim_name):
    if not getattr(obj, 'animation_data', None):
        return

    for track in obj.animation_data.nla_tracks:
        if track.name != anim_name:
            continue
        if not track.strips:
            continue

        obj.animation_data.action = track.strips[0].action
        return

    obj.animation_data.action = None

def make_fcurve(action, co, data_path, index=0, group_name='', interpolation=None):
    try:
        fcurve = action.fcurves.new(data_path=data_path, index=index, action_group=group_name)
    except:
        # Some non valid files can have multiple target path
        return None

    fcurve.keyframe_points.add(len(co) // 2)
    fcurve.keyframe_points.foreach_set('co', co)

    # Setting interpolation
    ipo = {
        'CUBICSPLINE': 'BEZIER',
        'LINEAR': 'LINEAR',
        'STEP': 'CONSTANT',
    }[interpolation or 'LINEAR']
    ipo = bpy.types.Keyframe.bl_rna.properties['interpolation'].enum_items[ipo].value
    fcurve.keyframe_points.foreach_set('interpolation', [ipo] * len(fcurve.keyframe_points))

    # For CUBICSPLINE, also set the handle types to AUTO
    if interpolation == 'CUBICSPLINE':
        ty = bpy.types.Keyframe.bl_rna.properties['handle_left_type'].enum_items['AUTO'].value
        fcurve.keyframe_points.foreach_set('handle_left_type', [ty] * len(fcurve.keyframe_points))
        fcurve.keyframe_points.foreach_set('handle_right_type', [ty] * len(fcurve.keyframe_points))

    fcurve.update() # force updating tangents (this may change when tangent will be managed)

    return fcurve



# Apache License, Version 2.0

# ./blender.bin --background -noaudio --python tests/python/bl_blendfile_library_overrides.py -- --output-dir=/tmp/
import pathlib
import bpy
import sys
import os
import unittest

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from bl_blendfile_utils import TestHelper


class TestLibraryOverrides(TestHelper, unittest.TestCase):
    MESH_LIBRARY_PARENT = "LibMeshParent"
    OBJECT_LIBRARY_PARENT = "LibMeshParent"
    MESH_LIBRARY_CHILD = "LibMeshChild"
    OBJECT_LIBRARY_CHILD = "LibMeshChild"
    MESH_LIBRARY_PERMISSIVE = "LibMeshPermissive"
    OBJECT_LIBRARY_PERMISSIVE = "LibMeshPermissive"

    def __init__(self, args):
        self.args = args

        output_dir = pathlib.Path(self.args.output_dir)
        self.ensure_path(str(output_dir))
        self.output_path = output_dir / "blendlib_overrides.blend"
        self.test_output_path = output_dir / "blendlib_overrides_test.blend"

        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        mesh = bpy.data.meshes.new(TestLibraryOverrides.MESH_LIBRARY_PARENT)
        obj = bpy.data.objects.new(TestLibraryOverrides.OBJECT_LIBRARY_PARENT, object_data=mesh)
        bpy.context.collection.objects.link(obj)
        mesh_child = bpy.data.meshes.new(TestLibraryOverrides.MESH_LIBRARY_CHILD)
        obj_child = bpy.data.objects.new(TestLibraryOverrides.OBJECT_LIBRARY_CHILD, object_data=mesh_child)
        obj_child.parent = obj
        bpy.context.collection.objects.link(obj_child)

        mesh = bpy.data.meshes.new(TestLibraryOverrides.MESH_LIBRARY_PERMISSIVE)
        obj = bpy.data.objects.new(TestLibraryOverrides.OBJECT_LIBRARY_PERMISSIVE, object_data=mesh)
        bpy.context.collection.objects.link(obj)
        obj.override_template_create()
        prop = obj.override_library.properties.add(rna_path='scale')
        prop.operations.add(operation='NOOP')

        bpy.ops.wm.save_as_mainfile(filepath=str(self.output_path), check_existing=False, compress=False)

    def __ensure_override_library_updated(self):
        # During save the override_library is updated.
        bpy.ops.wm.save_as_mainfile(filepath=str(self.test_output_path), check_existing=False, compress=False)

    def test_link_and_override_property(self):
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        bpy.data.orphans_purge()

        link_dir = self.output_path / "Object"
        bpy.ops.wm.link(directory=str(link_dir), filename=TestLibraryOverrides.OBJECT_LIBRARY_PARENT)

        obj = bpy.data.objects[TestLibraryOverrides.OBJECT_LIBRARY_PARENT]
        self.assertIsNone(obj.override_library)
        local_id = obj.override_create()
        self.assertIsNotNone(local_id.override_library)
        self.assertIsNone(local_id.data.override_library)
        assert(len(local_id.override_library.properties) == 0)

        local_id.location.y = 1.0

        self.__ensure_override_library_updated()

        assert(len(local_id.override_library.properties) == 1)
        override_prop = local_id.override_library.properties[0]
        assert(override_prop.rna_path == "location");
        assert(len(override_prop.operations) == 1)
        override_operation = override_prop.operations[0]
        assert(override_operation.operation == 'REPLACE')
        # Setting location.y overridded all elements in the location array. -1 is a wildcard.
        assert(override_operation.subitem_local_index == -1)
        
    def test_link_permissive(self):
        """
        Linked assets with a permissive template.

        - Checks if the NOOP is properly handled.
        - Checks if the correct properties and operations are created/updated.
        """
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        bpy.data.orphans_purge()

        link_dir = self.output_path / "Object"
        bpy.ops.wm.link(directory=str(link_dir), filename=TestLibraryOverrides.OBJECT_LIBRARY_PERMISSIVE)

        obj = bpy.data.objects[TestLibraryOverrides.OBJECT_LIBRARY_PERMISSIVE]
        self.assertIsNotNone(obj.override_library)
        local_id = obj.override_create()
        self.assertIsNotNone(local_id.override_library)
        self.assertIsNone(local_id.data.override_library)
        assert(len(local_id.override_library.properties) == 1)
        override_prop = local_id.override_library.properties[0]
        assert(override_prop.rna_path == "scale");
        assert(len(override_prop.operations) == 1)
        override_operation = override_prop.operations[0]
        assert(override_operation.operation == 'NOOP')
        assert(override_operation.subitem_local_index == -1)

        local_id.location.y = 1.0
        local_id.scale.x = 0.5
        # `scale.x` will apply, but will be reverted when the library overrides
        # are updated. This is by design so python scripts can still alter the
        # properties locally what is a typical usecase in productions.
        assert(local_id.scale.x == 0.5)
        assert(local_id.location.y == 1.0)

        self.__ensure_override_library_updated()
        assert(local_id.scale.x == 1.0)
        assert(local_id.location.y == 1.0)

        assert(len(local_id.override_library.properties) == 2)
        override_prop = local_id.override_library.properties[0]
        assert(override_prop.rna_path == "scale");
        assert(len(override_prop.operations) == 1)
        override_operation = override_prop.operations[0]
        assert(override_operation.operation == 'NOOP')
        assert(override_operation.subitem_local_index == -1)

        override_prop = local_id.override_library.properties[1]
        assert(override_prop.rna_path == "location");
        assert(len(override_prop.operations) == 1)
        override_operation = override_prop.operations[0]
        assert(override_operation.operation == 'REPLACE')
        assert (override_operation.subitem_local_index == -1)


class TestLibraryTemplate(TestHelper, unittest.TestCase):
    MESH_LIBRARY_PERMISSIVE = "LibMeshPermissive"
    OBJECT_LIBRARY_PERMISSIVE = "LibMeshPermissive"

    def __init__(self, args):
        pass

    def test_permissive_template(self):
        """
        Test setting up a permissive template.
        """
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        mesh = bpy.data.meshes.new(TestLibraryTemplate.MESH_LIBRARY_PERMISSIVE)
        obj = bpy.data.objects.new(TestLibraryTemplate.OBJECT_LIBRARY_PERMISSIVE, object_data=mesh)
        bpy.context.collection.objects.link(obj)
        assert(obj.override_library is None)
        obj.override_template_create()
        assert(obj.override_library is not None)
        assert(len(obj.override_library.properties) == 0)
        prop = obj.override_library.properties.add(rna_path='scale')
        assert(len(obj.override_library.properties) == 1)
        assert(len(prop.operations) == 0)
        operation = prop.operations.add(operation='NOOP')
        assert(len(prop.operations) == 1)
        assert(operation.operation == 'NOOP')


TESTS = (
    TestLibraryOverrides,
    TestLibraryTemplate,
)


def argparse_create():
    import argparse

    # When --help or no args are given, print this help
    description = "Test library overrides of blend file."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default=".",
        help="Where to output temp saved blendfiles",
        required=False,
    )

    return parser


def main():
    args = argparse_create().parse_args()

    # Don't write thumbnails into the home directory.
    bpy.context.preferences.filepaths.use_save_preview_images = False
    bpy.context.preferences.experimental.use_override_templates = True

    for Test in TESTS:
        Test(args).run_all_tests()


if __name__ == '__main__':
    import sys
    sys.argv = [__file__] + \
        (sys.argv[sys.argv.index("--") + 1:] if "--" in sys.argv else [])
    main()



#
# Copyright (C) 2016-2020 by Nathan Lovato, Daniel Oakey, Razvan Radulescu, and contributors
#
# This file is part of Power Sequencer.
#
# Power Sequencer is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Power Sequencer is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Power Sequencer. If
# not, see <https://www.gnu.org/licenses/>.
#
import bpy

from .utils.functions import convert_duration_to_frames
from .utils.global_settings import SequenceTypes
from .utils.doc import doc_name, doc_idname, doc_brief, doc_description


class POWER_SEQUENCER_OT_space_sequences(bpy.types.Operator):
    """
    *brief* Offsets all strips to the right of the time cursor by a given duration, ignoring locked sequences
    """

    doc = {
        "name": doc_name(__qualname__),
        "demo": "",
        "description": doc_description(__doc__),
        "shortcuts": [({"type": "EQUAL", "value": "PRESS"}, {}, "")],
        "keymap": "Sequencer",
    }
    bl_idname = doc_idname(__qualname__)
    bl_label = doc["name"]
    bl_description = doc_brief(doc["description"])
    bl_options = {"REGISTER", "UNDO"}

    gap_to_insert: bpy.props.FloatProperty(
        name="Duration", description="The time offset to apply to the strips", default=1.0
    )

    @classmethod
    def poll(cls, context):
        return context.sequences

    def invoke(self, context, event):
        sequences = [
            s
            for s in context.sequences
            if s.type in SequenceTypes.CUTABLE
            and s.frame_final_start >= context.scene.frame_current
            and not s.lock
        ]

        gap_frames = convert_duration_to_frames(context, self.gap_to_insert)
        sorted_sequences = sorted(sequences, key=lambda s: s.frame_final_start, reverse=True)
        for s in sorted_sequences:
            s.frame_start += gap_frames

        markers = context.scene.timeline_markers
        for m in [m for m in markers if m.frame >= context.scene.frame_current]:
            m.frame += gap_frames
        return {"FINISHED"}



# Purpose: acdsdata section manager
# Created: 05.05.2014
# Copyright (C) 2014, Manfred Moitzi
# License: MIT License

from __future__ import unicode_literals
__author__ = "mozman <mozman@gmx.at>"

from itertools import islice
from .tags import TagGroups, DXFStructureError, Tags, binary_encoded_data_to_bytes


class AcDsDataSection(object):
    name = 'acdsdata'

    def __init__(self):
        # Standard_ACIS_Binary (SAB) data store, key = handle of DXF Entity in the ENTITIES section: BODY, 3DSOLID
        # SURFACE, PLANESURFACE, REGION
        self.sab_data = {}

    @classmethod
    def from_tags(cls, tags, drawing):
        data_section = cls()
        data_section._build(tags)
        return data_section

    def _build(self, tags):
        if len(tags) == 3:  # empty entities section
            return

        for group in TagGroups(islice(tags, 2, len(tags)-1)):
            data_record = AcDsDataRecord(Tags(group))
            if data_record.dxftype == 'ACDSRECORD':
                asm_data = data_record.get_section('ASM_Data', None)
                if asm_data is not None:
                    self.add_asm_data(data_record)

    def add_asm_data(self, acdsrecord):
        """ Store SAB data as binary string in the sab_data dict, with handle to owner Entity as key.
        """
        try:
            asm_data = acdsrecord.get_section('ASM_Data')
            entity_id = acdsrecord.get_section('AcDbDs::ID')
        except ValueError:
            return
        else:
            handle = entity_id[2].value
            binary_data_text = (tag.value for tag in asm_data if tag.code == 310)
            binary_data = binary_encoded_data_to_bytes(binary_data_text)
            self.sab_data[handle] = binary_data


class Section(Tags):
    @property
    def name(self):
        return self[0].value

    @property
    def type(self):
        return self[1].value

    @property
    def data(self):
        return self[2:]


class AcDsDataRecord(object):
    def __init__(self, tags):
        self.dxftype = tags[0].value
        start_index = 2
        while tags[start_index].code != 2:
            start_index += 1
        self.sections = [Section(tags) for tags in TagGroups(islice(tags, start_index, None), split_code=2)]

    def has_section(self, name):
        return self.get_section(name, default=None) is not None

    def get_section(self, name, default=KeyError):
        for section in self.sections:
            if section.name == name:
                return section
        if default is KeyError:
            raise KeyError(name)
        else:
            return default

    def __getitem__(self, name):
        return self.get_section(name)



# This stub runs a python script relative to the currently open
# blend file, useful when editing scripts externally.

import bpy
import os

# Use your own script name here:
filename = "my_script.py"

filepath = os.path.join(os.path.dirname(bpy.data.filepath), filename)
global_namespace = {"__file__": filepath, "__name__": "__main__"}
with open(filepath, 'rb') as file:
    exec(compile(file.read(), filepath, 'exec'), global_namespace)



# ############################################################
# Importing - Same For All Render Layer Tests
# ############################################################

import unittest
import os
import sys

from view_layer_common import *


# ############################################################
# Testing
# ############################################################

class UnitTesting(ViewLayerTesting):
    def test_make_single_user(self):
        """
        Really basic test, just to check for crashes on basic files.
        """
        import bpy
        scene = bpy.context.scene
        master_collection = scene.master_collection
        view_layer = bpy.context.view_layer
        ob = bpy.context.object

        # clean up the scene a bit
        for o in (o for o in view_layer.objects if o != ob):
            view_layer.collections[0].collection.objects.unlink(o)

        for v in (v for v in scene.view_layers if v != view_layer):
            scene.view_layers.remove(v)

        while master_collection.collections:
            master_collection.collections.remove(
                master_collection.collections[0])

        view_layer.collections.link(master_collection)
        ob.select_set(True)

        # update depsgraph
        view_layer.update()

        # test itself
        bpy.ops.object.make_single_user(object=True)


# ############################################################
# Main - Same For All Render Layer Tests
# ############################################################

if __name__ == '__main__':
    UnitTesting._extra_arguments = setup_extra_arguments(__file__)
    unittest.main()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# run this script in the game engine.
# or on the command line with...
#  ./blender.bin --background -noaudio --python tests/python/bl_rst_completeness.py
'''
filepath = "/src/blender/tests/python/bl_rst_completeness.py"
exec(compile(open(filepath).read(), filepath, 'exec'))
'''

import os

THIS_DIR = os.path.dirname(__file__)
RST_DIR = os.path.normpath(os.path.join(THIS_DIR, "..", "..", "doc", "python_api", "rst"))

import sys
sys.path.append(THIS_DIR)

import rst_to_doctree_mini


# (file, module)
modules = (
    ("bgl.rst", "bgl", True),
    ("gpu.rst", "gpu", False),
)


def is_directive_pydata(filepath, directive):
    if directive.type in {"function", "method", "class", "attribute", "data"}:
        return True
    elif directive.type in {"module", "note", "warning", "code-block", "hlist", "seealso"}:
        return False
    elif directive.type == "literalinclude":  # TODO
        return False
    else:
        print(directive_to_str(filepath, directive), end=" ")
        print("unknown directive type %r" % directive.type)
        return False


def directive_to_str(filepath, directive):
    return "%s:%d:%d:" % (filepath, directive.line + 1, directive.indent)


def directive_members_dict(filepath, directive_members):
    return {directive.value_strip: directive for directive in directive_members
            if is_directive_pydata(filepath, directive)}


def module_validate(filepath, mod, mod_name, doctree, partial_ok):
    # RST member missing from MODULE ???
    for directive in doctree:
        # print(directive.type)
        if is_directive_pydata(filepath, directive):
            attr = directive.value_strip
            has_attr = hasattr(mod, attr)
            ok = False
            if not has_attr:
                # so we can have glNormal docs cover glNormal3f
                if partial_ok:
                    for s in dir(mod):
                        if s.startswith(attr):
                            ok = True
                            break

                if not ok:
                    print(directive_to_str(filepath, directive), end=" ")
                    print("rst contains non existing member %r" % attr)

            # if its a class, scan down the class...
            # print(directive.type)
            if has_attr:
                if directive.type == "class":
                    cls = getattr(mod, attr)
                    # print("directive:      ", directive)
                    for directive_child in directive.members:
                        # print("directive_child: ", directive_child)
                        if is_directive_pydata(filepath, directive_child):
                            attr_child = directive_child.value_strip
                            if attr_child not in cls.__dict__:
                                attr_id = "%s.%s" % (attr, attr_child)
                                print(directive_to_str(filepath, directive_child), end=" ")
                                print("rst contains non existing class member %r" % attr_id)

    # MODULE member missing from RST ???
    doctree_dict = directive_members_dict(filepath, doctree)
    for attr in dir(mod):
        if attr.startswith("_"):
            continue

        directive = doctree_dict.get(attr)
        if directive is None:
            print("module contains undocumented member %r from %r" % ("%s.%s" % (mod_name, attr), filepath))
        else:
            if directive.type == "class":
                directive_dict = directive_members_dict(filepath, directive.members)
                cls = getattr(mod, attr)
                for attr_child in cls.__dict__.keys():
                    if attr_child.startswith("_"):
                        continue
                    if attr_child not in directive_dict:
                        attr_id = "%s.%s.%s" % (mod_name, attr, attr_child), filepath
                        print("module contains undocumented member %r from %r" % attr_id)


def main():

    for filename, modname, partial_ok in modules:
        filepath = os.path.join(RST_DIR, filename)
        if not os.path.exists(filepath):
            raise Exception("%r not found" % filepath)

        doctree = rst_to_doctree_mini.parse_rst_py(filepath)
        __import__(modname)
        mod = sys.modules[modname]

        module_validate(filepath, mod, modname, doctree, partial_ok)


if __name__ == "__main__":
    main()



#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
"""
Node Templates - Vignette, Vector Blur

Add a set of nodes with one click, in this version I added a "Vignette"
as first example.

There is no official way to make a vignette, this is just my approach at
it. Templates: On the Compositor's header, "Template" pulldown. Or hit W.
Vignette: Adjust the size and position of the vignette with the Ellipse
Mask's X/Y and width, height values.
"""

import bpy
from amaranth.node_editor.templates.vectorblur import AMTH_NODE_OT_AddTemplateVectorBlur
from amaranth.node_editor.templates.vignette import AMTH_NODE_OT_AddTemplateVignette


KEYMAPS = list()


# Node Templates Menu
class AMTH_NODE_MT_amaranth_templates(bpy.types.Menu):
    bl_idname = 'AMTH_NODE_MT_amaranth_templates'
    bl_space_type = 'NODE_EDITOR'
    bl_label = "Templates"
    bl_description = "List of Amaranth Templates"

    def draw(self, context):
        layout = self.layout
        layout.operator(
            AMTH_NODE_OT_AddTemplateVectorBlur.bl_idname,
            text="Vector Blur",
            icon='FORCE_HARMONIC')
        layout.operator(
            AMTH_NODE_OT_AddTemplateVignette.bl_idname,
            text="Vignette",
            icon='COLOR')


def node_templates_pulldown(self, context):
    if context.space_data.tree_type == 'CompositorNodeTree':
        layout = self.layout
        row = layout.row(align=True)
        row.scale_x = 1.3
        row.menu("AMTH_NODE_MT_amaranth_templates",
                 icon="NODETREE")


def register():
    bpy.utils.register_class(AMTH_NODE_MT_amaranth_templates)
    bpy.utils.register_class(AMTH_NODE_OT_AddTemplateVignette)
    bpy.utils.register_class(AMTH_NODE_OT_AddTemplateVectorBlur)
    bpy.types.NODE_HT_header.append(node_templates_pulldown)
    kc = bpy.context.window_manager.keyconfigs.addon
    km = kc.keymaps.new(name="Node Editor", space_type="NODE_EDITOR")
    kmi = km.keymap_items.new("wm.call_menu", "W", "PRESS")
    kmi.properties.name = "AMTH_NODE_MT_amaranth_templates"
    KEYMAPS.append((km, kmi))


def unregister():
    bpy.utils.unregister_class(AMTH_NODE_MT_amaranth_templates)
    bpy.utils.unregister_class(AMTH_NODE_OT_AddTemplateVignette)
    bpy.utils.unregister_class(AMTH_NODE_OT_AddTemplateVectorBlur)
    bpy.types.NODE_HT_header.remove(node_templates_pulldown)
    for km, kmi in KEYMAPS:
        km.keymap_items.remove(kmi)
    KEYMAPS.clear()



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


from blenderkit import utils

import bpy
from object_print3d_utils import operators as ops

RENDER_OBTYPES = ['MESH', 'CURVE', 'SURFACE', 'METABALL', 'TEXT']


def check_material(props, mat):
    e = bpy.context.scene.render.engine
    shaders = []
    textures = []
    props.texture_count = 0
    props.node_count = 0
    props.total_megapixels = 0
    props.is_procedural = True

    if e == 'CYCLES':

        if mat.node_tree is not None:
            checknodes = mat.node_tree.nodes[:]
            while len(checknodes) > 0:
                n = checknodes.pop()
                props.node_count += 1
                if n.type == 'GROUP':  # dive deeper here.
                    checknodes.extend(n.node_tree.nodes)
                if len(n.outputs) == 1 and n.outputs[0].type == 'SHADER' and n.type != 'GROUP':
                    if n.type not in shaders:
                        shaders.append(n.type)
                if n.type == 'TEX_IMAGE':

                    if n.image is not None:
                        mattype = 'image based'
                        props.is_procedural = False
                        if n.image not in textures:
                            textures.append(n.image)
                            props.texture_count += 1
                            props.total_megapixels += (n.image.size[0] * n.image.size[1])

                            maxres = max(n.image.size[0], n.image.size[1])
                            props.texture_resolution_max = max(props.texture_resolution_max, maxres)
                            minres = min(n.image.size[0], n.image.size[1])
                            if props.texture_resolution_min == 0:
                                props.texture_resolution_min = minres
                            else:
                                props.texture_resolution_min = min(props.texture_resolution_min, minres)

    props.shaders = ''
    for s in shaders:
        if s.startswith('BSDF_'):
            s = s[5:]
        s = s.lower().replace('_', ' ')
        props.shaders += (s + ', ')


def check_render_engine(props, obs):
    ob = obs[0]
    m = None

    e = bpy.context.scene.render.engine
    mattype = None
    materials = []
    shaders = []
    textures = []
    props.uv = False
    props.texture_count = 0
    props.total_megapixels = 0
    props.node_count = 0
    for ob in obs:  # TODO , this is duplicated here for other engines, otherwise this should be more clever.
        for ms in ob.material_slots:
            if ms.material is not None:
                m = ms.material
                if m.name not in materials:
                    materials.append(m.name)
        if ob.type == 'MESH' and len(ob.data.uv_layers) > 0:
            props.uv = True

    if e == 'BLENDER_RENDER':
        props.engine = 'BLENDER_INTERNAL'
    elif e == 'CYCLES':

        props.engine = 'CYCLES'

        for mname in materials:
            m = bpy.data.materials[mname]
            if m is not None and m.node_tree is not None:
                checknodes = m.node_tree.nodes[:]
                while len(checknodes) > 0:
                    n = checknodes.pop()
                    props.node_count +=1
                    if n.type == 'GROUP':  # dive deeper here.
                        checknodes.extend(n.node_tree.nodes)
                    if len(n.outputs) == 1 and n.outputs[0].type == 'SHADER' and n.type != 'GROUP':
                        if n.type not in shaders:
                            shaders.append(n.type)
                    if n.type == 'TEX_IMAGE':


                        if n.image is not None and n.image not in textures:
                            props.is_procedural = False
                            mattype = 'image based'

                            textures.append(n.image)
                            props.texture_count += 1
                            props.total_megapixels += (n.image.size[0] * n.image.size[1])

                            maxres = max(n.image.size[0], n.image.size[1])
                            props.texture_resolution_max = max(props.texture_resolution_max, maxres)
                            minres = min(n.image.size[0], n.image.size[1])
                            if props.texture_resolution_min == 0:
                                props.texture_resolution_min = minres
                            else:
                                props.texture_resolution_min = min(props.texture_resolution_min, minres)


        # if mattype == None:
        #    mattype = 'procedural'
        # tags['material type'] = mattype

    elif e == 'BLENDER_GAME':
        props.engine = 'BLENDER_GAME'

    # write to object properties.
    props.materials = ''
    props.shaders = ''
    for m in materials:
        props.materials += (m + ', ')
    for s in shaders:
        if s.startswith('BSDF_'):
            s = s[5:]
        s = s.lower()
        s = s.replace('_', ' ')
        props.shaders += (s + ', ')


def check_printable(props, obs):
    if len(obs) == 1:
        check_cls = (
            ops.Print3DCheckSolid,
            ops.Print3DCheckIntersections,
            ops.Print3DCheckDegenerate,
            ops.Print3DCheckDistorted,
            ops.Print3DCheckThick,
            ops.Print3DCheckSharp,
            # ops.Print3DCheckOverhang,
        )

        ob = obs[0]

        info = []
        for cls in check_cls:
            cls.main_check(ob, info)

        printable = True
        for item in info:
            passed = item[0].endswith(' 0')
            if not passed:
                # print(item[0])
                printable = False

        props.printable_3d = printable


def check_rig(props, obs):
    for ob in obs:
        if ob.type == 'ARMATURE':
            props.rig = True


def check_anim(props, obs):
    animated = False
    for ob in obs:
        if ob.animation_data is not None:
            a = ob.animation_data.action
            if a is not None:
                for c in a.fcurves:
                    if len(c.keyframe_points) > 1:
                        animated = True

                        # c.keyframe_points.remove(c.keyframe_points[0])
    if animated:
        props.animated = True


def check_meshprops(props, obs):
    ''' checks polycount, manifold, mesh parts (not implemented)'''
    fc = 0
    fcr = 0
    tris = 0
    quads = 0
    ngons = 0
    vc = 0

    edges_counts = {}
    manifold = True

    for ob in obs:
        if ob.type == 'MESH' or ob.type == 'CURVE':
            ob_eval = None
            if ob.type == 'CURVE':
                # depsgraph = bpy.context.evaluated_depsgraph_get()
                # object_eval = ob.evaluated_get(depsgraph)
                mesh = ob.to_mesh()
            else:
                mesh = ob.data
            fco = len(mesh.polygons)
            fc += fco
            vc += len(mesh.vertices)
            fcor = fco
            for f in mesh.polygons:
                # face sides counter
                if len(f.vertices) == 3:
                    tris += 1
                elif len(f.vertices) == 4:
                    quads += 1
                elif len(f.vertices) > 4:
                    ngons += 1

                # manifold counter
                for i, v in enumerate(f.vertices):
                    v1 = f.vertices[i - 1]
                    e = (min(v, v1), max(v, v1))
                    edges_counts[e] = edges_counts.get(e, 0) + 1

            # all meshes have to be manifold for this to work.
            manifold = manifold and not any(i in edges_counts.values() for i in [0, 1, 3, 4])

            for m in ob.modifiers:
                if m.type == 'SUBSURF' or m.type == 'MULTIRES':
                    fcor *= 4 ** m.render_levels
                if m.type == 'SOLIDIFY':  # this is rough estimate, not to waste time with evaluating all nonmanifold edges
                    fcor *= 2
                if m.type == 'ARRAY':
                    fcor *= m.count
                if m.type == 'MIRROR':
                    fcor *= 2
                if m.type == 'DECIMATE':
                    fcor *= m.ratio
            fcr += fcor

            if ob_eval:
                ob_eval.to_mesh_clear()

    # write out props
    props.face_count = fc
    props.face_count_render = fcr
    # print(tris, quads, ngons)
    if quads > 0 and tris == 0 and ngons == 0:
        props.mesh_poly_type = 'QUAD'
    elif quads > tris and quads > ngons:
        props.mesh_poly_type = 'QUAD_DOMINANT'
    elif tris > quads and tris > quads:
        props.mesh_poly_type = 'TRI_DOMINANT'
    elif quads == 0 and tris > 0 and ngons == 0:
        props.mesh_poly_type = 'TRI'
    elif ngons > quads and ngons > tris:
        props.mesh_poly_type = 'NGON'
    else:
        props.mesh_poly_type = 'OTHER'

    props.manifold = manifold


def countObs(props, obs):
    ob_types = {}
    count = len(obs)
    for ob in obs:
        otype = ob.type.lower()
        ob_types[otype] = ob_types.get(otype, 0) + 1
    props.object_count = count


def check_modifiers(props, obs):
    # modif_mapping = {
    # }
    modifiers = []
    for ob in obs:
        for m in ob.modifiers:
            mtype = m.type
            mtype = mtype.replace('_', ' ')
            mtype = mtype.lower()
            # mtype = mtype.capitalize()
            if mtype not in modifiers:
                modifiers.append(mtype)
            if m.type == 'SMOKE':
                if m.smoke_type == 'FLOW':
                    smt = m.flow_settings.smoke_flow_type
                    if smt == 'BOTH' or smt == 'FIRE':
                        modifiers.append('fire')

    # for mt in modifiers:
    effectmodifiers = ['soft body', 'fluid simulation', 'particle system', 'collision', 'smoke', 'cloth',
                       'dynamic paint']
    for m in modifiers:
        if m in effectmodifiers:
            props.simulation = True
    if ob.rigid_body is not None:
        props.simulation = True
        modifiers.append('rigid body')
    finalstr = ''
    for m in modifiers:
        finalstr += m
        finalstr += ','
    props.modifiers = finalstr


def get_autotags():
    """ call all analysis functions """
    ui = bpy.context.scene.blenderkitUI
    if ui.asset_type == 'MODEL':
        ob = utils.get_active_model()
        obs = utils.get_hierarchy(ob)
        props = ob.blenderkit
        if props.name == "":
            props.name = ob.name

        # reset some properties here, because they might not get re-filled at all when they aren't needed anymore.
        props.texture_resolution_max = 0
        props.texture_resolution_min = 0
        # disabled printing checking, some 3d print addon bug.
        # check_printable( props, obs)
        check_render_engine(props, obs)

        dim, bbox_min, bbox_max = utils.get_dimensions(obs)
        props.dimensions = dim
        props.bbox_min = bbox_min
        props.bbox_max = bbox_max

        check_rig(props, obs)
        check_anim(props, obs)
        check_meshprops(props, obs)
        check_modifiers(props, obs)
        countObs(props, obs)
    elif ui.asset_type == 'MATERIAL':
        # reset some properties here, because they might not get re-filled at all when they aren't needed anymore.

        mat = utils.get_active_asset()
        props = mat.blenderkit
        props.texture_resolution_max = 0
        props.texture_resolution_min = 0
        check_material(props, mat)
    elif ui.asset_type == 'HDR':
        # reset some properties here, because they might not get re-filled at all when they aren't needed anymore.

        hdr = utils.get_active_asset()
        props = hdr.blenderkit
        props.texture_resolution_max = max(hdr.size[0],hdr.size[1])


class AutoFillTags(bpy.types.Operator):
    """Fill tags for asset. Now run before upload, no need to interact from user side"""
    bl_idname = "object.blenderkit_auto_tags"
    bl_label = "Generate Auto Tags for BlenderKit"
    bl_options = {'REGISTER', 'UNDO', 'INTERNAL'}

    @classmethod
    def poll(cls, context):
        return utils.uploadable_asset_poll()

    def execute(self, context):
        get_autotags()
        return {'FINISHED'}


def register_asset_inspector():
    bpy.utils.register_class(AutoFillTags)


def unregister_asset_inspector():
    bpy.utils.unregister_class(AutoFillTags)


if __name__ == "__main__":
    register()



#
# Copyright (C) 2016-2020 by Nathan Lovato, Daniel Oakey, Razvan Radulescu, and contributors
#
# This file is part of Power Sequencer.
#
# Power Sequencer is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Power Sequencer is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Power Sequencer. If
# not, see <https://www.gnu.org/licenses/>.
#
if __name__ == "__main__":
    import bpy

    render = bpy.context.scene.render
    render.resolution_x = 1920
    render.resolution_y = 1080
    render.resolution_percentage = 100
    render.pixel_aspect_x = 1
    render.pixel_aspect_y = 1

    render.image_settings.file_format = "FFMPEG"
    render.ffmpeg.format = "MPEG4"
    render.ffmpeg.codec = "H264"

    render.ffmpeg.constant_rate_factor = "PERC_LOSSLESS"
    render.ffmpeg.ffmpeg_preset = "BEST"

    scene = bpy.context.scene
    fps = scene.render.fps / scene.render.fps_base
    render.ffmpeg.gopsize = round(fps / 2.0)
    render.ffmpeg.use_max_b_frames = True
    render.ffmpeg.max_b_frames = 2

    render.ffmpeg.video_bitrate = 9000
    render.ffmpeg.maxrate = 9000
    render.ffmpeg.minrate = 0
    render.ffmpeg.buffersize = 224 * 8
    render.ffmpeg.packetsize = 2048
    render.ffmpeg.muxrate = 10080000

    render.ffmpeg.audio_codec = "AAC"
    render.ffmpeg.audio_bitrate = 384
